"""Abstract Metador container provider interface."""
from typing import Any, Dict, Generic, Optional, Protocol, Tuple, Type, TypeVar, Union

from .wrappers import MetadorContainer, MetadorDriver

T = TypeVar("T")


class ContainerProxy(Protocol[T]):
    """Abstract interface for Metador container providers.

    This interface acts like a proxy to access containers by some identifier.

    The identifier type parameter T is in the simplest case the Metador
    container UUID. In more complex cases, it could be a different unique
    identifier with a non-trivial relationship to Metador container UUIDs
    (many-to-many). Therefore, T is implementation-specific.

    There are many ways to store and organize containers, this interface serves
    as the implementation target for generic service components such as
    container-centric Flask blueprints, so they can be easier reused in
    different backends and services.

    Note that only containment and retrieval are possible - on purpose.
    Knowing and iterating over all containers in a system is not always possible.
    """

    def __contains__(self, key: T) -> bool:
        """Return whether a resource key is known to the proxy."""
        # return self.get(key) is not None

    def get(self, key: T) -> Optional[MetadorContainer]:
        """Get a container instance, if resource key is known to the proxy.

        Implement this method in subclasses to support the minimal interface.
        """

    def __getitem__(self, key: T) -> T:
        """Get a container instance, if resource key is known to the proxy.

        Default implementation is in terms of `get`.
        """
        if ret := self.get(key):
            return ret
        raise KeyError(key)


ContainerArgs = Tuple[Type[MetadorDriver], Any]
"""Pair of (driver class, suitable driver arguments).

Must be such that `MetadorContainer(driver(source))` yields a working container.
"""


class SimpleContainerProvider(Generic[T], ContainerProxy[T]):
    """Dict-backed container proxy.

    It is a minimal reasonable implementation for the interface that can be
    used in small apps and does not depend on the container driver,
    thus can support all container interface implementations.
    """

    _known: Dict[T, ContainerArgs]
    """Mapping from container identifier to MetadorContainer constructor args."""

    def __init__(self):
        self._known = {}

    def __contains__(self, key: T) -> bool:
        return key in self._known

    def get(self, key: T) -> Optional[MetadorContainer]:
        """Get an open container file to access data and metadata, if it exists."""
        if key not in self._known:
            return None
        driver, source = self._known[key]
        return MetadorContainer(driver(source))

    # ----

    def __delitem__(self, key: T):
        del self._known[key]

    def __setitem__(self, key: T, value: Union[ContainerArgs, MetadorContainer]):
        # NOTE: can't do instance check here, because MetadorContainer is a wrapper itself
        # so we check if a container is passed by presence of the .metador attribute
        if container_toc := getattr(value, "metador", None):
            self._known[key] = (container_toc.driver, container_toc.source)
        else:
            self._known[key] = value

    def keys(self):
        return self._known.keys()



from __future__ import annotations

import json
from typing import List, Optional, Union

import pytest
from pydantic import BaseModel

from metador_core.schema.jsonschema import (
    KEY_SCHEMA_HASH,
    clean_jsonschema,
    finalize_schema_extra,
    jsonschema_id,
    normalized_json,
    schema_of,
)


@pytest.mark.parametrize("x", [None, True, 10, 1.0, "hello"])
def test_clean_jsonschema_primitives(x):
    assert clean_jsonschema(x) == x


dirt = {
    "title": "some title",
    "description": "blah",
    "examples": ["a", "b"],
    "readOnly": True,
    "writeOnly": False,
    "deprecated": False,
    "$id": "some_identifier",
    "definitions": {"uiae": {}},
    "$defs": {"dtrn": 42},
    KEY_SCHEMA_HASH: "uiae",
}
"""Noise to be added to json schemas."""


schema = {
    "required": ["x"],
    "properties": {
        "x": {
            "type": "string",
            "format": "email",
        },
        "y": {
            "type": "int",
            "exclusiveMinimum": 10,
        },
    },
}

schema_dirty = {
    **dirt,  # type: ignore
    "required": ["x"],
    "properties": {
        "x": {
            **dirt,  # type: ignore
            "type": "string",
            "format": "email",
        },
        "y": {
            **dirt,  # type: ignore
            "type": "int",
            "exclusiveMinimum": 10,
        },
    },
}


def test_clean_jsonschema_nested():
    assert clean_jsonschema(schema_dirty) == schema


def test_normalized_json_str():
    input = {"y": None, "x": [True, 1.234e-10, "z", 567]}
    expected = b'{"x":[true,1.234e-10,"z",567],"y":null}'
    result = normalized_json(input)
    assert result == expected
    assert json.loads(result) == input


def test_jsonschema_id_sanity_check():
    assert jsonschema_id(schema) == jsonschema_id(schema_dirty)


# ----


class MyBaseModel(BaseModel):
    """Minimal base model with custom JSON Schema hacks."""

    class Config:
        @staticmethod
        def schema_extra(schema, model):
            print("call extra", model.__name__)
            finalize_schema_extra(schema, model, base_model=MyBaseModel)
            print("done extra", model.__name__)

    @classmethod
    def schema(cls, *args, **kwargs):
        print("call schema", cls.__name__)

        if cls.__base__ is MyBaseModel:
            ...  # no split inheritance
            # just move definitions -> $defs and rename models
        else:
            ...  # split inheritance, normalize
        ret = schema_of(cls, *args, **kwargs)

        print("done schema", cls.__name__)
        return ret


class Model1(MyBaseModel):
    val1: Optional[Model2]
    foo: Model4


class Model2(Model1):
    val2: Optional[List[Union[str, Model1]]]
    qux: bool


class Model3(MyBaseModel):
    bar: str
    baz: Union[int, str]


class Model4(Model3):
    baz: int
    quux: float


for m in [Model1, Model2, Model3, Model4]:
    m.update_forward_refs()

# Model1.schema()


@pytest.mark.skip(reason="FIXME")
def test_lift_nested_defs():
    s, h = [], []
    for m in [Model1, Model2, Model3, Model4]:
        s.append(dict(m.schema()))
        h.append(s[-1]["$ref"].split("/")[-1])
        # ensure we get same schema each time
        assert m.schema() == s[-1]
        # ensure the referenced schema is included
        assert h[-1] in s[-1]["$defs"]
        # check schema_json agrees with schema
        assert json.loads(m.schema_json()) == s[-1]

    # expected number of schemas in the dumps (one per (parent) model)
    assert list(map(lambda x: len(x["$defs"]), s)) == [4, 4, 1, 2]

    # same result (they cross-refer each other)
    assert s[0]["$defs"] == s[1]["$defs"]

    print(Model1.schema_json(indent=2))



"""Parse a SKOS ConceptScheme, explore it and generate enums for schemas.

Assumptions:
* ConceptScheme collects Concepts via hasTopConcept
* Concepts have 0-1 broader Concept and 0-n narrower Concept
* Concepts define a prefLabel
* Concepts have a unique IRI (ideally resolving to the concept sub-graph)

All of this applies to e.g. https://data.nist.gov/od/dm/nmrr/vocab
"""
from __future__ import annotations

from enum import Enum
from typing import List, Optional, Tuple, Type, cast

import rdflib
from rdflib.namespace import SKOS

from ..util import pythonize_name
from .lib import GraphNode, RDFParser


class SemanticEnum(Enum):
    """Enum subclass for Enums generated from a semantic taxonomy."""

    __self__term__: Tuple[str, str]


class Concept(RDFParser):
    """A concept is a node in a taxonomy defining a term."""

    _depth: int = 0
    """Depth of the concept in the taxonomy forest."""

    id: str
    """ID of the concept (should be IRI resolving to concept definition)."""

    prefLabel: str
    """Canonical name of the concept."""

    broader: Optional[Concept]
    """Unique more general concept in the taxonomy (unless it is a root)."""

    narrower: List[Concept]
    """A list of more specific sub-concepts."""

    def __eq__(self, other):
        """Return whether two concept objects are equal.

        For our purposes, concepts are the same if they come from the same graph
        and refer to the same IRI.
        """
        return (
            isinstance(other, Concept)
            and self.__wrapped__.graph == other.__wrapped__.graph
            and self.id == other.id
        )

    def new_subconcept(self, node: GraphNode):
        ret = Concept(node)
        ret._depth = self._depth + 1
        return ret

    def new_superconcept(self, node: GraphNode):
        ret = Concept(node)
        ret._depth = self._depth - 1
        return ret

    # ----

    def parse_id(self, node) -> str:
        assert isinstance(node, rdflib.URIRef)
        return node.toPython()

    def parse_prefLabel(self, node: GraphNode) -> str:
        val = node.object(SKOS.prefLabel)
        assert val.is_literal()
        return val.value

    def parse_broader(self, node: GraphNode) -> Optional[Concept]:
        if v := node.object(SKOS.broader):
            return self.new_superconcept(v)
        return None

    def parse_narrower(self, node: GraphNode) -> List[Concept]:
        q = node.objects(SKOS.narrower)
        return list(map(self.new_subconcept, q))

    # ----

    def pretty_print(
        self, *, max_depth=None, indent: Optional[int] = None, indent_unit: str = "\t"
    ):
        indent = indent or 0
        line = f"{indent*indent_unit}{self.prefLabel} -> {self.id}"
        lines = [line]

        if max_depth is None or max_depth > 0:
            max_depth_next = max_depth - 1 if max_depth else None
            lines += list(
                map(
                    lambda x: x.pretty_print(
                        indent=indent + 1, max_depth=max_depth_next
                    ),
                    self.narrower,
                )
            )

        return "\n".join(lines)

    def __str__(self):
        return self.pretty_print()

    # ----

    @property
    def term(self):
        """Return (ID, string) pair for this concept."""
        return (self.id, pythonize_name(self.prefLabel))

    def sub_terms(self, *, deep: bool = False):
        """Return dict of subconcepts (recursively, if desired)."""
        ret = dict(map(lambda x: x.term, self.narrower))
        if deep:
            _, pyname = self.term
            ret.update(
                dict(
                    t
                    for dct in map(lambda x: x.sub_terms(deep=deep), self.narrower)
                    for t in dct.items()
                )
            )
        return ret

    def to_enum(self, deep: bool = False) -> Type[SemanticEnum]:
        """Return Enum with immediate child concepts as possible values."""
        if deep:  # pragma: no cover
            # TODO: think how to combine the enums in the best way
            raise NotImplementedError

        ts = self.sub_terms(deep=deep)
        assert len(ts) == len(
            set(ts.values())
        )  # expect that human values are also unique
        ret = cast(
            Type[SemanticEnum],
            SemanticEnum(  # type: ignore
                f"{self.term[1].capitalize()}_Enum", {v: k for k, v in ts.items()}
            ),
        )
        # useful information for "deep" mode:
        ret.__self_term__ = self.term  # type: ignore
        return ret


class ConceptScheme(RDFParser):
    """A concept scheme points to the roots of a taxonomy forest.

    The the top level concepts are assumed to be unrelated
    (otherwise they should be united by the broader super-concept).

    For this reason, you cannot generate an enum on this level.
    """

    id: str
    hasTopConcept: List[Concept]

    def parse_id(self, node):
        assert isinstance(node, rdflib.URIRef)
        return node.toPython()

    def parse_hasTopConcept(self, node: GraphNode):
        q = node.objects(SKOS.hasTopConcept)
        return list(map(Concept, q))

    # ----

    def pretty_print(self, **kwargs):
        return "\n".join(map(lambda x: x.pretty_print(**kwargs), self.hasTopConcept))

    def __str__(self):
        return self.pretty_print()



import enum
from collections import ChainMap
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Set,
    Tuple,
    Type,
    Union,
)

import typing_extensions as te

# import typing_utils  # for issubtype
from runtype import validation as rv  # for is_subtype
from typing_extensions import Annotated, ClassVar, TypedDict

get_args = te.get_args  # re-export get_args

TypeHint: Any
"""For documentation purposes - to mark type hint arguments."""


def get_type_hints(cls) -> Mapping[str, Any]:
    """Return type hints of this class."""
    return te.get_type_hints(cls, include_extras=True)


def get_annotations(cls, *, all: bool = False) -> Mapping[str, Any]:
    """Return (non-inherited) annotations (unparsed) of given class."""
    if not all:
        return cls.__dict__.get("__annotations__", {})
    return ChainMap(*(c.__dict__.get("__annotations__", {}) for c in cls.__mro__))


to38hint: Dict[Type, Any] = {
    list: List,
    set: Set,
    dict: Dict,
    type: Type,
}
"""Type hint map for consistent behavior between python versions."""


def get_origin(hint):
    # Fixed get_origin to return 3.8-compatible type hints.
    # Need this for pydantic dynamic model generation in 3.8.
    o = te.get_origin(hint)
    return to38hint.get(o, o)


# ----


def is_list(hint):
    return get_origin(hint) is List


def is_set(hint):
    return get_origin(hint) is Set


def is_union(hint):
    return get_origin(hint) is Union


def is_classvar(hint):
    return get_origin(hint) is ClassVar


def is_annotated(hint):
    return get_origin(hint) is Annotated


def is_literal(hint):
    return get_origin(hint) is Literal


NoneType = type(None)


def is_nonetype(hint):
    return hint is NoneType


def is_optional(hint):
    # internally, Optional is just sugar for a Union including NoneType.
    return is_union(hint) and any(map(is_nonetype, get_args(hint)))


# ----


def make_typehint(h, *args):
    if is_optional(h):
        args_ = list(args)
        args_.append(type(None))
    elif is_annotated(h):
        args_ = (args[0],)
    else:
        args_ = args
    return h.copy_with(tuple(args_))


UNION_PROXY = Union[int, str]

LIT = Literal[1]
TUP = Tuple[Any]


def make_literal(val):
    """Given a JSON object, return type that parses exactly that object.

    Note that dicts can have extra fields that will be ignored
    and that coercion between bool and int might happen.

    Sets and floats are not supported.
    """
    if val is None:
        return type(None)
    elif isinstance(val, (bool, int, str)):
        return make_typehint(LIT, val)
    elif issubclass(val.__class__, enum.Enum):
        return make_typehint(LIT, val.value)
    elif isinstance(val, (tuple, list)):
        args = tuple(map(make_literal, val))
        return make_typehint(TUP, *args)
    elif isinstance(val, dict):
        d = {k: make_literal(v) for k, v in val.items()}
        # NOTE: the TypedDict must be from typing_extensions for 3.8!
        return TypedDict("AnonConstDict", d)  # type: ignore
    raise ValueError(f"Unsupported value: {val}")


def make_tree_traversal(succ_func: Callable[[Any], Iterable]):
    """Return generator to traverse nodes of a tree-shaped object.

    Returned function has a boolean keyword argument `post_order`.
    If True, will emit the parent node after children instead of before.

    Args:
        succ_func: Function to be called on each node returning Iterable of children
    """

    def traverse(obj, *, post_order: bool = False):
        if not post_order:
            yield obj
        for t in succ_func(obj):
            yield from traverse(t, post_order=post_order)
        if post_order:
            yield obj

    return traverse


traverse_typehint = make_tree_traversal(get_args)
"""Perform depth-first pre-order traversal of a type annotation.

Args:
    th (object): type hint object to be traversed
"""


def make_tree_mapper(node_constructor, succ_func):
    def map_func(obj, leaf_map_func):
        if children := succ_func(obj):
            mcs = (map_func(c, leaf_map_func) for c in children)
            return node_constructor(obj, *mcs)
        else:
            return leaf_map_func(obj)

    return map_func


map_typehint = make_tree_mapper(make_typehint, get_args)


def unoptional(th):
    """Return type hint that is not optional (if it was optional)."""
    if is_annotated(th):
        # remove inner optional, preserve annotation
        return make_typehint(th, unoptional(get_args(th)[0]))

    if not is_union(th):
        # all optionals are actually unions -> nothing to do
        return th

    # filter out NoneType from the Union arguments
    args = tuple(filter(lambda h: not is_nonetype(h), get_args(th)))
    if len(args) == 1:
        # not a union anymore -> remove type
        return args[0]
    # remove union without NoneType (i.e. not optional)
    return make_typehint(UNION_PROXY, *args)


# ----


def is_subtype(sub, base):
    # add hack to ignore pydantic Annotated FieldInfo
    # add hacky fix for literals
    # NOTE: this is only superficial, actually issubtype must be fixed
    # or it won't work with nested Annotated types or complicated stuff
    ann_sub, ann_base = is_annotated(sub), is_annotated(base)
    lit_sub, lit_base = is_literal(sub), is_literal(base)
    if ann_sub != ann_base or lit_sub != lit_base:
        return False  # not equal on annotated wrapping status

    if not ann_sub:
        # proceed as usual
        return rv.is_subtype(sub, base)
        # return typing_utils.issubtype(sub, base)
    else:
        sub_args, base_args = get_args(sub), get_args(base)
        # NOTE: FieldInfo of pydantic is not comparable :( so we ignore it
        # same_ann = list(sub_args)[1:] == list(base_args)[1:]
        return is_subtype(sub_args[0], base_args[0])  # and same_ann


def is_subtype_of(t: Any) -> Callable[[Any], bool]:
    """Return a predicate to check issubtype for a given type."""
    return lambda obj: is_subtype(obj, t)


def is_instance_of(t: Any) -> Callable[[Any], bool]:
    """Return a predicate to check isinstance for a given type."""
    return lambda obj: isinstance(obj, t)


def is_subclass_of(t: Any) -> Callable[[Any], bool]:
    """Return a predicate to check issubclass for a given type."""
    return lambda obj: isinstance(obj, type) and issubclass(obj, t)


is_enum = is_subclass_of(enum.Enum)



"""Schema plugin group."""

from __future__ import annotations

from typing import Dict, List, Optional, Set, Type

from ..plugin import interface as pg
from ..plugins import plugingroups
from .core import MetadataSchema, check_types, infer_parent
from .partial import PartialModel
from .plugins import PluginBase
from .plugins import PluginRef as AnyPluginRef
from .types import SemVerTuple

SCHEMA_GROUP_NAME = "schema"  # name of schema plugin group


class SchemaPlugin(PluginBase):
    """Schema-specific Plugin section."""

    auxiliary: bool = False
    """If set to True, the schema is considered auxiliary.

    The consequence is that metadata objects based on this schema cannot be
    attached to containers.

    The intended for schema plugins that are too general or not self-contained,
    but could be useful in a larger context, e.g. as a parent schema or nested
    schema.
    """


class PGSchema(pg.PluginGroup[MetadataSchema]):
    """Interface to access installed schema plugins.

    All registered schema plugins can be used anywhere in a Metador container to
    annotate any group or dataset with metadata objects following that schema.

    If you don't want that, do not register the schema as a plugin, but just use
    the schema class as a normal Python dependency. Schemas that are not
    registered as plugins still must inherit from MetadataSchema, to ensure that
    all required methods are available and work as expected by the system.

    Unregistered schemas can be used as "abstract" parent schemas that cannot be
    instantiated in containers because they are too general to be useful, or for
    schemas that are not intended to be used on their own in the container, but
    model a meaningful metadata object that can be part of a larger schema.


    Guidelines for field definition:

    * Stick to the following types to construct your field annotation:
      - basic types: (`bool, int, float, str`)
      - basic hints from `typing`: `Optional, Literal, Union, Set, List, Tuple`
      - default pydantic types (such as `AnyHttpUrl`)
      - default classes supported by pydantic (e.g. `enum.Enum`, `datetime`, etc.)
      - constrained types defined using the `phantom` package
      - valid schema classes (subclasses of `MetadataSchema`)

    * `Optional` is for values that are semantically *missing*,
      You must not assume that a `None` value represents anything else than that.

    * Prefer `Set` over `List` when order is irrelevant and duplicates are not needed

    * Avoid using plain `Dict`, always define a schema instead if you know the keys,
      unless you really need to "pass through" whatever is given, which is usually
      not necessary for schemas that you design from scratch.

    * Prefer types from `phantom` over using pydantic `Field` settings for expressing
      simple value constraints (e.g. minimal/maximal value or collection length, etc.),
      because `phantom` types can be subclassed to narrow them down.

    * In general, avoid using `Field` at all, except for defining an `alias` for
      attributes that are not valid as Python variables (e.g. `@id` or `$schema`).

    * When using `Field`, make sure to annotate it with `typing_extensions.Annotated`,
      instead of assigning the `Field` object to the field name.


    Rules for schema versioning:

    All schemas must be direct or indirect subclass of `MetadataSchema`.

    Semantic versioning (MAJOR, MINOR, PATCH) is to be followed.
    Bumping a version component means incrementing it and resetting the
    later ones to 0. When updating a schema, you must bump:

    * PATCH, if you do not modify the set of parsable instances,

    * MINOR, if if your changes strictly increase parsable instances,

    * MAJOR otherwise, i.e. some older metadata might not be valid anymore.

    If you update a nested or inherited schema to a version
    with higher X (MAJOR, MINOR or PATCH), the version
    of your schema must be bumped in X as well.


    Rules for schema subclassing:

    A child schema that only extends a parent with new fields is safe.
    To schemas that redefine parent fields additional rules apply:

    EACH instance of a schema MUST also be parsable by the parent schema

    This means that a child schema may only override parent fields
    with more specific types, i.e., only RESTRICT the set of acceptable
    values compared to the parent field (safe examples include
    adding new or narrowing existing bounds and constraints,
    or excluding some values that are allowed by the parent schema).

    As automatically verifying this in full generality is not feasible, but the
    ability to "restrict" fields is very much needed in practical use, the
    schema developer MUST create suitable represantative test cases that check
    whether this property is satisfied.

    Try expressing field value restrictions by:

    * removing alternatives from a `Union`
    * using a subclass of a schema or `phantom` type that was used in the parent

    These can be automatically checked most of the time.
    """

    class Plugin:
        name = SCHEMA_GROUP_NAME
        version = (0, 1, 0)
        requires = [plugingroups.Plugin.ref()]
        plugin_class = MetadataSchema
        plugin_info_class = SchemaPlugin

    def __post_init__(self):
        self._parent_schema: Dict[
            Type[MetadataSchema], Optional[Type[MetadataSchema]]
        ] = {}
        self._parents: Dict[AnyPluginRef, List[AnyPluginRef]] = {}  # base plugins
        self._children: Dict[AnyPluginRef, Set[AnyPluginRef]] = {}  # subclass plugins

        # used schemas inside schemas
        self._field_types: Dict[
            Type[MetadataSchema], Dict[str, Set[Type[MetadataSchema]]]
        ] = {}
        self._subschemas: Dict[MetadataSchema, Set[MetadataSchema]] = {}

        # partial schema classes
        self._partials: Dict[MetadataSchema, PartialModel] = {}
        self._forwardrefs: Dict[str, MetadataSchema] = {}

    def plugin_deps(self, plugin) -> Set[AnyPluginRef]:
        self._parent_schema[plugin] = infer_parent(plugin)
        if pcls := self._parent_schema[plugin]:
            # make sure a parent schema plugin is initialized before the child
            info = pcls.Plugin
            return {self.PluginRef(name=info.name, version=info.version)}
        else:
            return set()

    def check_plugin(self, name: str, plugin: Type[MetadataSchema]):
        check_types(plugin)  # ensure that (overrides of) fields are valid

    def _compute_parent_path(self, plugin: Type[MetadataSchema]) -> List[AnyPluginRef]:
        ref = plugin.Plugin.ref()
        ret = [ref]
        curr = plugin
        parent = self._parent_schema[curr]
        while parent is not None:
            p_ref = parent.Plugin.ref()
            ret.append(p_ref)
            curr = self._get_unsafe(p_ref.name, p_ref.version)
            parent = self._parent_schema[curr]

        ret.reverse()
        return ret

    def init_plugin(self, plugin):
        # pre-compute parent schema path
        ref = plugin.Plugin.ref()
        self._parents[ref] = self._compute_parent_path(plugin)
        if ref not in self._children:
            self._children[ref] = set()

        # collect children schema set for all parents
        parents = self._parents[ref][:-1]
        for parent in parents:
            self._children[parent].add(ref)

    # ----

    def parent_path(
        self, schema, version: Optional[SemVerTuple] = None
    ) -> List[AnyPluginRef]:
        """Get sequence of registered parent schema plugins leading to the given schema.

        This sequence can be a subset of the parent sequences in the actual class
        hierarchy (not every subclass must be registered as a plugin).
        """
        name, vers = pg.plugin_args(schema, version, require_version=True)
        ref = self.PluginRef(name=name, version=vers)
        self._ensure_is_loaded(ref)
        return list(self._parents[ref])

    def children(
        self, schema, version: Optional[SemVerTuple] = None
    ) -> Set[AnyPluginRef]:
        """Get set of names of registered (strict) child schemas."""
        name, vers = pg.plugin_args(schema, version, require_version=True)
        ref = self.PluginRef(name=name, version=vers)
        self._ensure_is_loaded(ref)
        return set(self._children[ref])


SchemaPlugin.update_forward_refs()



import importlib_metadata
import pytest

from metador_core.plugin.metaclass import UndefVersion
from metador_core.plugin.types import is_metador_ep_group, to_ep_group_name, to_ep_name
from metador_core.plugin.util import register_in_group
from metador_core.plugins import plugingroups

from . import dummy_plugins as d

# basic tests


def test_plugingroup_misc(plugingroups_test):
    plugingroups = plugingroups_test
    schemas = plugingroups_test["schema"]

    # check name
    assert plugingroups.name == "plugingroup"
    assert schemas.name == "schema"

    # check packages
    assert plugingroups.packages == schemas.packages
    assert "metador-core" in plugingroups.packages

    assert plugingroups.provider(schemas.Plugin.ref()).name == "metador-core"
    assert schemas.provider(schemas["core.file"].Plugin.ref()).name == "metador-core"


def test_plugingroup_str_repr(plugingroups_test):
    # just make sure it works without error and dummy check output
    schemas = plugingroups_test["schema"]
    rstr = repr(schemas)
    assert rstr.find("core.file")
    pstr = str(schemas)
    assert pstr.find("core.file")
    assert pstr.find("\n")


def test_plugingroup_pluginref(plugingroups_test):
    """Check that PluginRef subclasses work as expected."""
    # check PluginRef of the plugingroups themselves
    plugingroups = plugingroups_test
    schemas = plugingroups_test["schema"]

    pg_ref = plugingroups.PluginRef(name="plugingroup", version=(0, 1, 0))
    assert pg_ref.group == pg_ref.name
    assert plugingroups.Plugin.ref() == pg_ref

    sg_ref = schemas.Plugin.ref()
    assert sg_ref.group == pg_ref.name
    assert sg_ref.name == "schema"

    # check a plugin PluginRef
    s_ref = next(schemas.keys())
    assert s_ref.group == sg_ref.name


def test_is_plugin(plugingroups_test):
    plugingroups = plugingroups_test
    schemas = plugingroups_test["schema"]

    assert plugingroups.is_plugin(schemas)
    assert not schemas.is_plugin(plugingroups)

    PCls = next(iter(schemas.values()))

    Marked = schemas.get(PCls.Plugin.name)
    assert UndefVersion._is_marked(Marked)

    class Impostor(PCls):
        ...

    Impostor.Plugin = PCls.Plugin  # metaclass removes it, add back for test

    class Impostor2(PCls):
        class Plugin:
            name = PCls.Plugin.name
            version = PCls.Plugin.version

    class Impostor3(PCls):
        ...

    delattr(Impostor3, "Plugin")

    assert PCls in schemas
    assert Marked in schemas
    # impostor class can pretend to be a plugin in many contexts
    assert Impostor in schemas
    assert Impostor2 in schemas

    assert schemas.is_plugin(PCls)
    assert schemas.is_plugin(Marked)
    # but when we really want to check, we can find out
    assert not schemas.is_plugin(Impostor)
    assert not schemas.is_plugin(Impostor2)
    assert not schemas.is_plugin(Impostor3)


def test_plugingroup_compare(plugingroups_test):
    # different PluginRef subclasses still comparable!
    plugingroups = plugingroups_test
    schemas = plugingroups_test["schema"]

    from metador_core.schema.plugins import PluginRef

    kwargs = dict(name="xyz", version=(0, 1, 0))

    p_ref = plugingroups.PluginRef(**kwargs)
    s_ref = schemas.PluginRef(**kwargs)
    assert p_ref != s_ref
    assert p_ref == PluginRef(group=plugingroups.name, **kwargs)
    assert s_ref == PluginRef(group=schemas.name, **kwargs)


def test_plugingroup_dictlike_basic(plugingroups_test):
    """Do some manual sanity checks."""
    plugingroups = plugingroups_test
    schemas = plugingroups_test["schema"]

    # NOTE: important to check that plugingroup itself is a plugingroup
    # because it allows us to do the next rigorous checks easily

    # check __contains__
    assert plugingroups.Plugin.ref() in plugingroups
    assert schemas.Plugin.ref() in plugingroups
    assert "schema" in plugingroups

    # check __getitem__
    assert plugingroups["plugingroup"] is plugingroups
    assert plugingroups["schema"] is schemas
    assert plugingroups[schemas] is schemas  # type: ignore

    # check get
    assert plugingroups.get("plugingroup", (0, 1, 0)) is plugingroups
    assert plugingroups.get(plugingroups, (0, 1, 0)) is plugingroups
    assert plugingroups.get("schema") is schemas


@pytest.mark.parametrize("pg", list(plugingroups.values()))
def test_plugingroup_dictlike_consistency(pg):
    """Check consistency of dict-like operations on all (real) plugingroups."""
    # check keys, values, items, __contains__, __getitem__ and get
    # also check is_plugin
    pg_keys = set(pg.keys())
    pg_values = set(pg.values())
    for k, v in pg.items():
        # minimal invariant
        assert k in pg_keys
        assert v in pg_values
        assert pg.is_plugin(v)

        # lookup and access via name, ref or class
        v_ref = v.Plugin.ref()
        assert k in pg
        assert v_ref in pg
        assert v in pg

        assert pg[k] == v
        assert pg[v] == v
        assert pg[v_ref] == v

        assert pg.get(k) == v

    # now try something that should not exist
    invalid = "_invalid_plugin"
    assert not pg.is_plugin(invalid)
    assert invalid not in pg
    assert invalid not in pg_keys
    assert pg.get(invalid) is None
    with pytest.raises(KeyError):
        pg[invalid]


# try creating invalid plugin groups
@pytest.mark.parametrize("pg", [d.PGInvalid1, d.PGInvalid2, d.PGInvalid3, d.PGInvalid4])
def test_invalid_plugingroups(pg, plugingroups_test):
    with pytest.raises(TypeError):
        register_in_group(plugingroups_test, pg, violently=True)


# ----
# Test plugingroup behavior (adding, checking, loading plugins, multiple versions, etc)


@pytest.fixture
def pg_dummy(plugingroups_test):
    plugingroups_test.__reset__()  # plugingroups_test is module-level, need to reset it
    # add valid dummy plugin group into test plugin environment
    register_in_group(plugingroups_test, d.PGDummy, violently=True)
    # return to tests of plugin system
    return plugingroups_test.get("dummy")


class FakeDist:
    name = "foo"
    version = "1.2.3"


@pytest.fixture(scope="session")
def ep_factory():
    """Entry point maker."""

    def maker(obj, ep_name=None, group=None):
        if ep_name is None:
            ep_name = to_ep_name(obj.Plugin.name, obj.Plugin.version)

        assert not is_metador_ep_group(group)
        group = to_ep_group_name(group)
        return importlib_metadata.EntryPoint(
            ep_name, f"{obj.__module__}:{obj.__qualname__}", group
        )._for(FakeDist)

    return maker


def test_add_ep_invalid(pg_dummy, ep_factory):
    with pytest.raises(ValueError):
        pg_dummy._add_ep("invalid_entrypoint", None)


def test_add_ep_twice(plugingroups_test):
    schemas = plugingroups_test["schema"]

    plugin = schemas.get("core.file")
    ref = plugin.Plugin.ref()
    ep_name = to_ep_name(ref.name, ref.version)
    ep = schemas._ENTRY_POINTS[ep_name]

    ep2 = importlib_metadata.EntryPoint(ep.name, ep.value, ep.group)._for(FakeDist)
    schemas._add_ep(ep_name, ep2)
    assert schemas._ENTRY_POINTS[ep_name] == ep2


# ----


# try adding invalid plugins
def test_plugingroups(pg_dummy, ep_factory):
    assert not list(pg_dummy.keys())
    make_ep = ep_factory

    ep1 = make_ep(d.InvalidPlugin1, group=pg_dummy.name, ep_name="test.plugin__0.1.0")
    with pytest.raises(TypeError):
        pg_dummy._add_ep(ep1.name, ep1)
        pg_dummy["test.plugin"]  # no plugin inner class

    ep2 = make_ep(d.InvalidPlugin2, group=pg_dummy.name, ep_name="xyz__0.1.0")
    with pytest.raises(ValueError):
        pg_dummy._add_ep(ep2.name, ep2)
        pg_dummy._get_unsafe("xyz")  # mismatch ep name / stated name
    ep2b = make_ep(d.InvalidPlugin2, group=pg_dummy.name, ep_name="abc__0.2.0")
    with pytest.raises(ValueError):
        pg_dummy._add_ep(ep2b.name, ep2b)
        pg_dummy._get_unsafe("abc")  # mismatch ep version /stated version
    ep2c = make_ep(d.InvalidPlugin2b, group=pg_dummy.name, ep_name="test__0.1.0")
    with pytest.raises(ValueError):
        pg_dummy._add_ep(ep2c.name, ep2c)
        pg_dummy._get_unsafe("test")  # missing prefix

    ep3 = make_ep(d.InvalidPlugin3, group=pg_dummy.name)
    with pytest.raises(TypeError):
        pg_dummy._add_ep(ep3.name, ep3)
        pg_dummy._get_unsafe("test.plugin")  # missing 'something'

    ep4 = make_ep(d.InvalidPlugin4, group=pg_dummy.name)
    with pytest.raises(TypeError):
        pg_dummy._add_ep(ep4.name, ep4)
        pg_dummy._get_unsafe("test.plugin")  # not DummyBase subclass

    ep5 = make_ep(d.InvalidPlugin5, group=pg_dummy.name)
    with pytest.raises(TypeError):
        pg_dummy._add_ep(ep5.name, ep5)
        pg_dummy._get_unsafe("test.plugin")  # missing method

    # now this works
    ep6 = make_ep(d.DummyPlugin, group=pg_dummy.name)
    pg_dummy._add_ep(ep6.name, ep6)
    pg_dummy._get_unsafe("test.plugin")


# TODO: test multi version etc
def test_versions(pg_dummy, ep_factory):
    register_in_group(pg_dummy, d.DummyPlugin, violently=True)  # 0.1
    register_in_group(pg_dummy, d.DummyPluginB, violently=True)  # 0.1.1
    register_in_group(pg_dummy, d.DummyPluginC, violently=True)  # 1.2.3
    register_in_group(pg_dummy, d.DummyPlugin2, violently=True)

    # both plugins are available
    assert "test.plugin" in pg_dummy
    assert "test.plugin2" in pg_dummy

    # all plugins in all versions are listed
    keys = list(pg_dummy.keys())
    for p in [d.DummyPlugin, d.DummyPlugin2, d.DummyPluginB, d.DummyPluginC]:
        assert p.Plugin.ref() in keys
        assert p in pg_dummy

    # we can get all versions of a plugin
    expected = [(0, 1, 0), (0, 1, 1), (1, 2, 3)]
    vs = pg_dummy.versions("test.plugin")
    assert list(map(lambda x: x.version, vs)) == expected

    # access without version yields latest
    latest = pg_dummy.get("test.plugin")
    assert UndefVersion._is_marked(latest)
    assert UndefVersion._unwrap(latest) is d.DummyPluginC

    # access with version yields newest compatible
    # specifically higher minor is returned
    compat = pg_dummy.get("test.plugin", (0, 1, 0))
    assert not UndefVersion._is_marked(compat)
    assert compat is d.DummyPluginB
    assert pg_dummy.get("test.plugin", (1, 1, 0)) is d.DummyPluginC
    assert pg_dummy.get("test.plugin2", (0, 0, 1)) is d.DummyPlugin2

    # smaller "revision" may be returned (semantically equivalent)
    # TODO: is this actually a good idea?...
    assert pg_dummy.get("test.plugin", (0, 1, 2)) is d.DummyPluginB
    assert pg_dummy.get("test.plugin", (1, 2, 4)) is d.DummyPluginC
    assert pg_dummy.get("test.plugin2", (0, 1, 1)) is d.DummyPlugin2

    # these should not be found
    assert pg_dummy.get("test.plugin", (0, 2, 0)) is None
    assert pg_dummy.get("test.plugin2", (0, 2, 0)) is None
    assert pg_dummy.get("test.plugin", (2, 0, 0)) is None
    assert pg_dummy.get("test.plugin2", (1, 0, 0)) is None


plugingroups.__reset__()  # because we accessed the real plugin groups



import pytest
from pydantic import ValidationError

from metador_core.schema.plugins import PluginBase, PluginRef


def test_pluginref():
    with pytest.raises(ValidationError):
        # extra fields forbidden
        PluginRef(group="a", name="b", version=(0, 1, 0), xtra="haha")

    ref = PluginRef(group="a", name="b", version=(0, 1, 0))
    ref2 = PluginRef(group="mygroup", name="b", version=(0, 1, 0))

    with pytest.raises(TypeError):
        ref.group = "x"  # immutable

    assert len(str(ref).split("\n")) == 1  # str without indentation (wasteful)

    # test subclass
    MyPluginRef = PluginRef._subclass_for("mygroup")
    mref = MyPluginRef(name="b", version=(0, 1, 0))

    with pytest.raises(ValidationError):
        # cannot override group for subclass
        MyPluginRef(**ref.dict())

    # hash works as expected (class does not matter, only by value)
    s = set()
    s.add(mref)
    assert mref in s
    assert ref2 in s

    # (in)equality also works between subclasses
    assert mref != ref
    assert mref == MyPluginRef(name="b", version=(0, 1, 0))
    assert mref == ref2

    # comparison works
    assert ref < PluginRef(group="b", name="a", version=(1, 0, 0))  # first by group
    assert mref < MyPluginRef(name="c", version=(0, 0, 1))  # by name
    assert mref < MyPluginRef(name="b", version=(0, 1, 1))  # by version

    # test "supports"
    assert not ref.supports(ref2)  # other group
    assert not ref.supports(ref.copy(update=dict(name="y")))  # other name
    assert ref.supports(ref.copy(update=dict(version=(0, 0, 1))))
    assert ref.supports(ref.copy(update=dict(version=(0, 1, 1))))
    assert not ref.supports(ref.copy(update=dict(version=(0, 2, 0))))
    assert not ref.supports(ref.copy(update=dict(version=(1, 0, 0))))


def test_pluginbase(plugingroups_test):
    class DummyPlugin(PluginBase):
        ...

    DummyPlugin.group = "schema"

    # test invalid
    with pytest.raises(TypeError):

        class DummyInfoFail1:
            name = "dummy"
            version = "1.2.3"  # <- invalid type

        DummyPlugin.parse_info(DummyInfoFail1)

    with pytest.raises(ValueError):
        # mismatch with EP name
        class DummyInfoFail2:
            name = "dummy"
            version = (1, 2, 3)

        DummyPlugin.parse_info(DummyInfoFail2, ep_name="dummy__4.5.6")

    # test valid
    class DummyInfo:
        name = "dummy"
        version = (1, 2, 3)

    info = DummyPlugin.parse_info(DummyInfo, ep_name="dummy__1.2.3")
    info = DummyPlugin.parse_info(DummyInfo)
    assert info.ref() == PluginRef(group="schema", name="dummy", version=(1, 2, 3))
    assert info.ref(version=(4, 5, 6)) == PluginRef(
        group="schema", name="dummy", version=(4, 5, 6)
    )

    assert info.plugin_string() == "metador.schema.dummy.1.2.3"

    # pretty printed JSON str
    assert len(str(info).split("\n")) > 1



import secrets


def random_hex(length: int) -> str:
    """Return random hex string of given length."""
    return secrets.token_hex(length // 2 + 1)[:length]


def parameters(d, keys=None):
    """Expand parameter combinations from a nested dict into a list of tuples."""
    keys = keys or []
    if isinstance(d, list):
        return sum((parameters(y, keys) for y in d), [])
    elif isinstance(d, dict):
        return sum((parameters(v, keys + [k]) for k, v in d.items()), [])
    else:
        return [tuple(keys + [d])]



from collections import ChainMap
from dataclasses import dataclass
from io import UnsupportedOperation
from typing import Any, Callable, List, Mapping, Optional, Set, Type

import wrapt
from pydantic import BaseModel
from simple_parsing.docstring import get_attribute_docstring

from ..util.models import field_origins
from ..util.typing import get_annotations


class LiftedRODict(type):
    """Metaclass for classes providing dict keys as attributes.

    Mostly for aesthetic reasons and to be used for things
    where the dict is actually a fixed lookup table.

    We don't provide explicit `keys`/`values`/`items`, because
    these could be key names in the dict.

    You can use `iter` to go through the keys and use dict-like
    access, if dynamic iteration is needed.
    """

    # NOTE: we don't want to add non-default methods
    # because method names collide with dict keys

    _dict: Mapping[str, Any]
    """The underlying dict."""

    _keys: Optional[List[str]] = None
    """Optionally, list of keys in desired order."""

    _repr: Optional[Callable] = None
    """Optional custom repr string or function."""

    def __repr__(self):
        # choose best representation based on configuration
        if self._repr:
            return self._repr(self)
        if self._keys:
            return repr(self._keys)
        return repr(list(self._dict.keys()))

    def __dir__(self):
        # helpful for tab completion
        return list(self._dict.keys())

    def __bool__(self):
        return bool(self._dict)

    def __contains__(self, key):
        return key in self._dict

    def __iter__(self):
        if self._keys:
            return iter(self._keys)
        return iter(self._dict)

    def __getitem__(self, key):
        return self._dict[key]

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError as e:
            raise AttributeError(str(e))

    def __setattr__(self, key, value):
        # this is supposed to be read-only
        raise UnsupportedOperation


def lift_dict(name, dct, *, keys=None, repr=None):
    """Return LiftedRODict class based on passed dict."""
    assert hasattr(dct, "__getitem__")
    kwargs = {"_dict": dct}
    if keys is not None:
        assert set(keys) == set(iter(dct))
        kwargs["_keys"] = keys
    if repr is not None:
        kwargs["_repr"] = repr
    return LiftedRODict(name, (), kwargs)


class WrappedLiftedDict(wrapt.ObjectProxy):
    """Wrap values returned by a LiftedRODict."""

    def __init__(self, obj, wrapperfun):
        assert isinstance(obj, LiftedRODict)
        super().__init__(obj)
        self._self_wrapperfun = wrapperfun

    def __getitem__(self, key):
        return self._self_wrapperfun(self.__wrapped__[key])

    def __getattr__(self, key):
        return LiftedRODict.__getattr__(self, key)

    def __repr__(self):
        return repr(self.__wrapped__)


@dataclass
class FieldInspector:
    """Basic field inspector carrying type and description of a field."""

    origin: Type
    name: str
    type: str

    description: str  # declared for proper repr generation

    def _get_description(self):
        desc = get_attribute_docstring(self.origin, self.name).docstring_below
        if not desc:
            # if none set, try getting docstring from field type
            if ths := getattr(self.origin, "_typehints", None):
                th = ths[self.name]
                if isinstance(th, type):  # it's a class-like thing?
                    desc = th.__doc__

        return desc

    @property  # type: ignore
    def description(self):
        # look up on-demand and cache, could be expensive (parses source)
        if not hasattr(self, "_description"):
            self._description = self._get_description()
        return self._description

    def __init__(self, model: Type[BaseModel], name: str, hint: str):
        origin = next(field_origins(model, name))
        self.origin = origin
        self.name = name
        self.type = hint


def make_field_inspector(
    model: Type[BaseModel],
    prop_name: str,
    *,
    bound: Optional[Type[BaseModel]] = BaseModel,
    key_filter: Optional[Callable[[str], bool]],
    i_cls: Optional[Type[FieldInspector]] = FieldInspector,
) -> Type[LiftedRODict]:
    """Create a field inspector class for the given model.

    This can be used for introspection about fields and also
    enables users to access subschemas without extra imports,
    improving decoupling of plugins and packages.

    To be used in a metaclass for a custom top level model.

    Args:
        model: Class for which to return the inspector
        prop_name: Name of the metaclass property that wraps this function
        i_cls: Optional subclass of FieldInspector to customize it
        bound: Top level class using the custom metaclass that uses this function
        key_filter: Predicate used to filter the annotations that are to be inspectable
    Returns:
        A fresh inspector class for the fields.
    """
    # get hints corresponding to fields that are not inherited
    field_hints = {
        k: v
        for k, v in get_annotations(model).items()
        if not key_filter or key_filter(k)
    }

    # inspectors for fields declared in the given model (for inherited, will reuse/create parent inspectors)
    new_inspectors = {k: i_cls(model, k, v) for k, v in field_hints.items()}
    # manually compute desired traversal order (from newest overwritten to oldest inherited fields)
    # as the default chain map order semantically is not suitable.
    inspectors = [new_inspectors] + [
        getattr(b, prop_name)._dict for b in model.__bases__ if issubclass(b, bound)
    ]
    covered_keys: Set[str] = set()
    ordered_keys: List[str] = []
    for d in inspectors:
        rem_keys = set(iter(d)) - covered_keys
        covered_keys.update(rem_keys)
        ordered_keys += [k for k in d if k in rem_keys]

    # construct and return the class
    return lift_dict(
        f"{model.__name__}.{prop_name}",
        ChainMap(*inspectors),
        keys=ordered_keys,
        repr=lambda self: "\n".join(map(str, (self[k] for k in self))),
    )



"""Test plain IH5 record."""
from pathlib import Path
from uuid import uuid1

import h5py
import numpy as np
import pytest

from metador_core.ih5.container import IH5Record, IH5UserBlock


def test_raw_open_empty_record():
    # list of files must be non-empty
    with pytest.raises(ValueError):
        IH5Record._open([])


def test_open_create_invalid_record_name(tmp_ds_path):
    # record names: only alphanumeric and dashes allowed
    for name in ["invalid.name", "invalid_name", "Юникод", "inva#lid"]:
        assert not IH5Record._is_valid_record_name(name)
        with pytest.raises(ValueError):
            IH5Record(tmp_ds_path / name)
        with pytest.raises(ValueError):
            IH5Record(tmp_ds_path / name, "w")


def test_open_r(tmp_ds_path):
    # not existing yet -> cannot open
    assert IH5Record.find_files(tmp_ds_path) == []

    with pytest.raises(FileNotFoundError):
        IH5Record(tmp_ds_path)
    with IH5Record(tmp_ds_path, "w"):
        pass
    with IH5Record(tmp_ds_path) as ds:
        assert len(ds.ih5_files) == 1
        assert ds.mode == "r"
        with pytest.raises(ValueError):
            ds["test"] = "write"


def test_open_rplus(tmp_ds_path):
    # not existing yet -> cannot open
    assert IH5Record.find_files(tmp_ds_path) == []
    with pytest.raises(FileNotFoundError):
        IH5Record(tmp_ds_path)
    with IH5Record(tmp_ds_path, "w"):
        pass
    with IH5Record(tmp_ds_path, "r+") as ds:
        assert len(ds.ih5_files) == 2
        assert ds.mode == "r+"
        ds["test"] = "write"  # no error


def test_open_a(tmp_ds_path):
    # missing -> should be created
    assert IH5Record.find_files(tmp_ds_path) == []
    with IH5Record(tmp_ds_path, "a") as ds:
        assert ds.mode == "r+"
        ds["test"] = "write"  # writable
    # creates a new patch for existing one
    with IH5Record(tmp_ds_path, "a") as ds:
        assert ds.mode == "r+"
        assert len(ds.ih5_files) == 2
        assert ds["test"][()] == b"write"
        ds["test2"] = "write_more"  # writable


def test_open_w(tmp_ds_path):
    # Open with "w" will kill _all_ connected files if a record exists
    with IH5Record(tmp_ds_path, "w") as ds:
        for _ in range(3):
            ds.commit_patch()
            ds.create_patch()
        old_ds_uuid = ds.ih5_uuid

    with IH5Record(tmp_ds_path, "w") as ds:
        pass

    with IH5Record(tmp_ds_path) as ds:
        assert ds.ih5_uuid != old_ds_uuid
        assert len(ds.ih5_files) == 1


def test_open_x(tmp_ds_path):
    # x/w- should not overwrite existing
    with IH5Record(tmp_ds_path, "w"):
        pass
    with pytest.raises(FileExistsError):
        IH5Record(tmp_ds_path, "x")
    with pytest.raises(FileExistsError):
        IH5Record(tmp_ds_path, "w-")


def test_create_open(tmp_ds_path):
    # create a container, add some stuff, do some sanity checks
    ds = IH5Record(tmp_ds_path, "w")

    # check that the expected container path is created and properties are set
    assert len(ds.ih5_files) == 1
    assert len(IH5Record.find_files(tmp_ds_path)) == 1
    assert str(ds.ih5_files[0]).find(str(tmp_ds_path)) == 0
    assert ds._infer_name(tmp_ds_path) == tmp_ds_path.name
    assert ds.ih5_uuid == ds.ih5_meta[0].record_uuid

    # add some data
    ds["foo"] = "bar"
    del ds["foo"]
    ds["blub"] = "bla"
    ds.create_group("grp")
    ds.attrs["qux"] = 123  # type: ignore
    ds.close()

    with IH5Record(tmp_ds_path) as ds2:
        assert "blub" in ds2
        assert "grp" in ds2
        assert "something" not in ds2
        assert "qux" in ds2.attrs

        assert ds2["blub"][()] == b"bla"
        assert ds2.attrs["qux"] == 123

        assert set(ds2.keys()) == set(["blub", "grp"])
        assert set(iter(ds2)) == set(["blub", "grp"])
        assert set(ds2.values()) == set(map(lambda x: ds2[x], ds2.keys()))
        assert set(map(lambda x: x[1], ds2.items())) == set(ds2.values())

        # we're read-only now, so these should fail:
        with pytest.raises(ValueError):
            ds2["key"] = 123  # setters should not work
        with pytest.raises(ValueError):
            del ds2["blub"]  # deleting should not work
        with pytest.raises(ValueError):
            ds2.commit_patch()
        with pytest.raises(ValueError):
            ds2.discard_patch()


def test_missing_userblock_open_fail(tmp_ds_path):
    # try loading with a missing userblock
    with IH5Record(tmp_ds_path, "w") as ds:
        cfile = ds.ih5_files[0]
    f = h5py.File(cfile, "w")
    f.close()

    with pytest.raises(ValueError):
        IH5Record(tmp_ds_path)


def test_broken_userblock_open_fail(tmp_ds_path):
    # try loading with an invalid userblock
    with IH5Record(tmp_ds_path, "w") as ds:
        cfile = ds.ih5_files[0]
    with open(cfile, "r+b") as f:
        f.write(b"broken")  # destroy the header

    with pytest.raises(ValueError):
        IH5Record(tmp_ds_path)


def test_larger_userblock_open(tmp_ds_path):
    # try loading a user block with bigger than default size
    # (for upward compatibility of the format, if the block needs to grow)
    with IH5Record(tmp_ds_path, "w") as ds:
        cfile = ds.ih5_files[0]
    with h5py.File(cfile, mode="w", userblock_size=2048):
        pass
    ub = IH5UserBlock.create()
    ub._userblock_size = 2048
    ub.save(cfile)

    rub = IH5UserBlock.load(cfile)
    assert rub == ub
    assert rub._userblock_size == 2048


def test_userblock_missing_save_fail(tmp_ds_path):
    # saving userblock should fail in HDF5 files that have no space reserved
    with IH5Record(tmp_ds_path, "w") as ds:
        cfile = ds.ih5_files[0]
    with h5py.File(cfile, mode="w"):
        pass
    ub = IH5UserBlock.create()

    with pytest.raises(ValueError):
        ub.save(cfile)


def test_create_patch_discard(tmp_ds_path):
    # try creating a patch and later discard it
    # also try calling these methods multiple times when it is not valid
    with IH5Record(tmp_ds_path, "w") as ds:
        assert len(ds.ih5_files) == 1
        assert ds._has_writable
        ds["foo"] = 123

        with pytest.raises(ValueError):  # cannot discard base container
            ds.discard_patch()

        with pytest.raises(ValueError):  # cannot create, have writable
            ds.create_patch()

        ds.commit_patch()  # commit base container
        assert not ds._has_writable

        with pytest.raises(ValueError):  # after commit nothing to discard
            ds.discard_patch()

        assert len(ds.ih5_files) == 1

        assert "bar" not in ds
        with pytest.raises(ValueError):  # should not work yet
            ds["bar"] = 456

        ds.create_patch()
        assert ds._has_writable
        with pytest.raises(ValueError):  # should not work twice
            ds.create_patch()
        assert len(ds.ih5_files) == 2  # new container was created

        ds["bar"] = 456
        assert "bar" in ds

        ds.discard_patch()
        assert not ds._has_writable

        assert len(ds.ih5_files) == 1
        assert "bar" not in ds

    with IH5Record(tmp_ds_path) as ds:  # re-open, look again
        assert len(ds.ih5_files) == 1
        assert "foo" in ds
        assert "bar" not in ds

        with pytest.raises(ValueError):  # cannot commit, read-only
            ds.commit_patch()


def test_create_patch_commit_patch(tmp_ds_path):
    with IH5Record(tmp_ds_path, "w") as ds:
        assert len(ds.ih5_files) == 1
        ds["foo"] = 123
        ds["foo"].attrs["qux"] = 321  # type: ignore
        ds.commit_patch()
        ds.create_patch()
        assert len(ds.ih5_files) == 2
        ds["bar"] = 456
        ds["foo"].attrs["qux"] = 789  # type: ignore
        ds.commit_patch()
        ds.create_patch()
        assert len(ds.ih5_files) == 3
        del ds["bar"]
        ds["bar"] = 1337
        ds.commit_patch()

    with IH5Record(tmp_ds_path) as ds:  # re-open, look again
        assert len(ds.ih5_files) == 3
        assert "foo" in ds
        assert "bar" in ds
        assert ds["foo"].attrs["qux"] == 789
        assert ds["bar"][()] == 1337


def test_open_scrambled_filenames(tmp_ds_path):
    # try opening files where the filenames are not systematic and out of order
    with IH5Record(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        for _ in range(3):
            ds.create_patch()
            ds.commit_patch()
        files = ds.ih5_files
        uuids = [ds.ih5_meta[i].patch_uuid for i in range(4)]

    # permutate the files
    order = [3, 0, 2, 1]
    newfiles = [Path(""), Path(""), Path(""), Path("")]
    for i in range(4):
        newfiles[order[i]] = files[i].parent / f"scramble{order[i]}"
        files[i].rename(newfiles[order[i]])

    # check that the files were re-ordered back into the correct order
    with IH5Record._open(newfiles) as ds:
        for i in range(4):
            assert ds.ih5_meta[i].patch_uuid == uuids[i]


def test_check_baseless_fileset_open(tmp_ds_path):
    # try opening an incomplete (patches only) set of files forming a patch chain
    # (not many reasons to do it, but technically there is nothing special about a base)
    with IH5Record(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        for _ in range(3):
            ds.create_patch()
            ds.commit_patch()
        paths = ds.ih5_files  # get container files. we care about the patches only

    with pytest.raises(ValueError):  # with base, but invalid
        IH5Record._open([paths[0], paths[2]])
    with pytest.raises(ValueError):  # baseless, but it is forbidden
        IH5Record._open(paths[1:])
    with pytest.raises(ValueError):  # baseless, but invalid
        IH5Record._open([paths[1], paths[3]], allow_baseless=True)
    with IH5Record._open(paths[1:], allow_baseless=True) as ds:  # baseless, and allowed
        pass


def test_create_patch_then_merge(tmp_ds_path_factory):
    dsname, target = tmp_ds_path_factory(), tmp_ds_path_factory()
    with IH5Record(dsname, "w") as ds:
        ds.attrs["bool_attr"] = True
        ds["foo"] = 123
        ds.commit_patch()
        ds.create_patch()
        ds["bar"] = [1, 2, 3]
        ds["bar"].attrs["str_attr"] = "something"  # type: ignore
        ds.commit_patch()
        ds.create_patch()
        ds.create_group("baz")
        ds["baz"].attrs["int_attr"] = 42
        ds["qux/dat"] = np.void(b"somedata")

        with pytest.raises(ValueError):  # not read-only -> calling merge shall fail
            ds.merge_files(target)

        ds.commit_patch()

        # now merge record and compare contents to original
        merged_file = ds.merge_files(target)
        # merged_file = merged.ih5_files[-1]
        # print(merged_file)
        # merged.close()
        with IH5Record(target) as ds2:
            # check user block
            assert ds2.ih5_uuid == ds.ih5_uuid
            assert len(ds2.ih5_files) == 1
            assert ds2._ublock(0).patch_index == ds._ublock(-1).patch_index
            assert ds2._ublock(0).patch_uuid == ds._ublock(-1).patch_uuid
            assert ds2._ublock(0).prev_patch is None
            assert ds2._ublock(0).hdf5_hashsum.find("sha256:") == 0

            # check data
            orig_nodes, copy_nodes = [], []
            ds.visit(orig_nodes.append)
            ds2.visit(copy_nodes.append)
            assert orig_nodes == copy_nodes
            assert ds.attrs["bool_attr"] == ds2.attrs["bool_attr"]
            assert ds["foo"][()] == ds2["foo"][()]  # type: ignore
            assert np.array_equal(ds["bar"][()], ds2["bar"][()])  # type: ignore
            assert ds["bar"].attrs["str_attr"] == ds2["bar"].attrs["str_attr"]  # type: ignore
            assert ds["baz"].attrs["int_attr"] == ds2["baz"].attrs["int_attr"]
            assert ds["qux/dat"][()].tobytes() == ds2["qux/dat"][()].tobytes()  # type: ignore

    # now create another patch for the original record
    # and then try to open it with the merged container
    with IH5Record(dsname, "r+") as ds:
        ds["qux/new_entry"] = "amazing data"
        assert "qux/new_entry" in ds
        del ds.attrs["bool_attr"]
        del ds["foo"]
        ds["foo"] = 456
        ds.commit_patch()
        new_patch = ds.ih5_files[-1]

    # we should be able to open the merged file with the new patch
    with IH5Record._open([merged_file, new_patch]) as ds:
        assert "bool_attr" not in ds.attrs
        assert ds["foo"][()] == 456
        assert "qux/dat" in ds
        assert "qux/new_entry" in ds
        assert ds["qux/new_entry"][()] == b"amazing data"  # type: ignore


def test_clear_empty(tmp_ds_path):
    # A cleared out multi-patch container is recognized as empty correctly.
    def is_empty(ds):
        return not len(ds) and not len(ds.attrs)

    with IH5Record(tmp_ds_path, "w") as ds:
        assert is_empty(ds)
        ds.attrs["atr"] = "value"
        assert not is_empty(ds)
        ds._clear()  # clear "for real"
        assert is_empty(ds)

        ds.attrs["atr"] = "value"
        ds.commit_patch()
        ds.create_patch()
        ds["foo"] = "bar"
        ds.commit_patch()
        ds.create_patch()
        assert not is_empty(ds)
        ds._clear()  # clear in overlay
        assert is_empty(ds)


def test_delete_record(tmp_ds_path):
    # Deleting closes and removes all files.
    with IH5Record(tmp_ds_path, "w") as ds:
        for _ in range(3):
            ds.commit_patch()
            ds.create_patch()
        files = ds.ih5_files
    assert ds._closed

    IH5Record.delete_files(tmp_ds_path)
    for file in files:
        assert not file.is_file()


def test_get(tmp_ds_path):
    # dict-like get method with default fallback value
    with IH5Record(tmp_ds_path, "w") as ds:
        assert ds.get("foo") is None
        assert ds.get("foo", 123) == 123
        ds["foo"] = 456
        assert ds.get("foo")[()] == 456


def test_create_dataset(tmp_ds_path):
    # relative and absolute paths for create_dataset + pass through arguments
    with IH5Record(tmp_ds_path, "w") as ds:
        foo = ds.create_group("foo")
        foo.create_dataset(
            "bar", data=[1, 2, 3], compression="gzip", compression_opts=9
        )
        foo.create_dataset("/baz", data=1)
        ds.create_dataset("qux/quux", data=2)
        assert "bar" in foo
        assert "foo/bar" in ds
        assert "/baz" in ds
        assert "/qux/quux" in ds

        # try invalid arguments
        with pytest.raises(ValueError) as e:
            ds.create_dataset("newdataset", unknown="arg")
        assert str(e).lower().find("unknown")


# --------
# Test opening containers with various failures
# by manipulating data into invalid states and
# checking that exceptions are correctly triggered


def test_check_ublock_inconsistent_dsuuid_fail(tmp_ds_path):
    # make that there are multiple different record uuids in the files
    with IH5Record(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        ds.create_patch()
        ds._ublock(-1).record_uuid = uuid1()
        ds.commit_patch()

    with pytest.raises(ValueError):
        IH5Record(tmp_ds_path)


def test_check_ublock_inconsistent_checksum_fail(tmp_ds_path):
    # make that checksum for data is incorrect
    with IH5Record(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        ds.create_patch()
        ds.commit_patch()
        ub = ds._ublock(-1)
        ub.hdf5_hashsum = (
            "sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
        )
        ub.save(ds._files[-1].filename)

    with pytest.raises(ValueError):
        IH5Record(tmp_ds_path)


def test_check_ublock_base_with_prev_patch_fail(tmp_ds_path):
    # make that previous patch uuid does not match
    with IH5Record(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        ds.create_patch()
        ds._ublock(-1).prev_patch = None
        ds.commit_patch()

    with pytest.raises(ValueError):
        IH5Record(tmp_ds_path)


def test_check_ublock_patch_without_prev_patch_fail(tmp_ds_path):
    # make that a patch does not have a declared predecessor
    with IH5Record(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        ds.create_patch()
        ds._ublock(-1).prev_patch = None
        ds.commit_patch()

    with pytest.raises(ValueError):
        IH5Record(tmp_ds_path)


def test_check_ublock_base_patch_idx_order_fail(tmp_ds_path):
    # try loading with files in invalid patch index order
    with IH5Record(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        ds.create_patch()
        ds.commit_patch()
        ub = ds._ublock(0)
        ub.patch_index = 123
        ub.save(ds._files[0].filename)

        with pytest.raises(ValueError):
            ds._check_ublock(ds._files[1].filename, ds._ublock(1), ds._ublock(0))


def test_check_ublock_base_patch_uuid_mismatch_fail(tmp_ds_path):
    # make that the linked predecessor uuid does not match with actual predecessor
    with IH5Record(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        ds.create_patch()
        ds._ublock(1).prev_patch = uuid1()
        ds.commit_patch()

    with pytest.raises(ValueError):
        IH5Record(tmp_ds_path)


def test_check_ublock_multiple_same_patch_uuid_fail(tmp_ds_path):
    # make that there are multiple files with the same patch UUID
    with IH5Record(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        ds.create_patch()
        ds._ublock(1).patch_uuid = ds._ublock(0).patch_uuid
        ds.commit_patch()

    with pytest.raises(ValueError):
        IH5Record(tmp_ds_path)


def test_not_open_fail(tmp_ds_path_factory):
    ds = IH5Record(tmp_ds_path_factory(), "w")
    ds.close()

    def assert_ex(f):
        with pytest.raises(ValueError) as e:
            f()
        assert str(e).lower().find("not open") >= 0

    # check that public methods fail gracefully when record not open
    assert_ex(lambda: ds.create_patch())
    assert_ex(lambda: ds.discard_patch())
    assert_ex(lambda: ds.commit_patch())
    assert_ex(lambda: ds.merge_files(tmp_ds_path_factory()))



"""Useful types and validators for use in pydantic models."""
from __future__ import annotations

import isodate
from phantom.re import FullMatch
from pint import Quantity, UndefinedUnitError, Unit
from pydantic import StrictBool, StrictBytes, StrictFloat, StrictInt, StrictStr
from typing_extensions import TypeAlias

from ..plugin.types import SemVerTuple, from_semver_str, to_semver_str
from ..util.hashsums import _hash_alg
from .encoder import json_encoder
from .parser import BaseParser, ParserMixin

# ----

# we want people to use the strict types, they should be default.
# coercing / normalization should be done explicitly!
# before, we needed to use fixes from https://github.com/pydantic/pydantic/issues/2329
# with pydantic 1.10.2 we can just pass through to the strict types

Bool: TypeAlias = StrictBool
Int: TypeAlias = StrictInt
Float: TypeAlias = StrictFloat
Str: TypeAlias = StrictStr
Bytes: TypeAlias = StrictBytes

# ----

# we prefer this over pydantic anystr config settings (non-local) and
# we use that instead of NonEmpty[str] because we can also react to whitespace
# and we can subclass it:


class NonEmptyStr(FullMatch, pattern=r"\s*\S[\S\s]*"):
    """Non-empty string (contains non-whitespace characters)."""


class MimeTypeStr(NonEmptyStr, pattern=r"[^ /;]+/[^ /;]+(;[^ /;]+)*"):
    """String that looks like a mime-type."""


class HashsumStr(NonEmptyStr, pattern="[0-9a-fA-F]+"):
    """String that looks like a hashsum."""


_hashalg_regex = f"(?:{'|'.join(_hash_alg.keys())})"


class QualHashsumStr(HashsumStr, pattern=_hashalg_regex + r":[0-9a-fA-F]+"):
    """Hashsum string, prepended by the used algorithm."""


# ----


@json_encoder(isodate.duration_isoformat)
class Duration(ParserMixin, isodate.Duration):
    """ISO 8601 Duration."""

    class Parser(BaseParser):
        schema_info = dict(
            title="string in ISO 8601 duration format",
            type="string",
            examples=["PT3H4M1S"],
        )

        @classmethod
        def parse(cls, tcls, v):
            if not isinstance(v, (str, tcls)):
                raise TypeError(f"Expected str or Duration, got {type(v)}.")
            # we have to force it into a Duration object,
            # otherwise we get possibly a timedelta back, which we do not want
            # because we want to serialize to the ISO format in both cases
            dur = isodate.parse_duration(v) if isinstance(v, str) else v
            return tcls(seconds=dur.total_seconds())


# Physical units


class StringParser(BaseParser):
    """Parser from string into some target class."""

    @classmethod
    def parse(cls, tcls, v):
        if isinstance(v, tcls):
            return v

        if not isinstance(v, str):
            msg = f"Expected str or {tcls.__name__}, got {type(v)}."
            raise TypeError(msg)

        ret = tcls(v)
        return ret


class PintParser(StringParser):
    """Shared base for `PintUnit` and `PintQuantity`, taking care of exceptions."""

    @classmethod
    def parse(cls, tcls, v):
        if not v:
            msg = f"Got empty string, expected {tcls.__name__}."
            raise ValueError(msg)
        try:
            return super().parse(tcls, v)
        except UndefinedUnitError as e:
            raise ValueError(str(e))


@json_encoder(str)
class PintUnit(ParserMixin, Unit):
    """pydantic-compatible pint.Unit."""

    class Parser(PintParser):
        schema_info = dict(
            title="Physical unit compatible with the Python pint library.",
            type="string",
            examples=["meter * candela", "kilogram / second ** 2"],
        )


@json_encoder(str)
class PintQuantity(ParserMixin, Quantity):
    """pydantic-compatible pint.Quantity."""

    def __new__(cls, *args, **kwargs):
        # hack to make parsing work, for some reason it does not work without this
        # (it does not correctly identify the unit for some reason)
        if kwargs.get("passthrough"):
            return super().__new__(cls, *args)

        ret = Quantity(*args)  # ensure that the quantity is correctly parsed
        # return instance of the subclass:
        return cls(ret.m, ret.u, passthrough=True)

    class Parser(PintParser):
        schema_info = dict(
            title="Physical quantity compatible with the Python pint library.",
            type="string",
            examples=["5 meter * candela", "7.12 kilogram / second ** 2"],
        )


__all__ = [
    "SemVerTuple",
    "to_semver_str",
    "from_semver_str",
    "Bool",
    "Int",
    "Float",
    "Bytes",
    "Str",
    "NonEmptyStr",
    "MimeTypeStr",
    "HashsumStr",
    "QualHashsumStr",
    "Duration",
    "StringParser",
    "PintParser",
    "PintUnit",
    "PintQuantity",
]



"""Immutable HDF5 container records.

A record consists of a base container and a number of patch containers.
This allows a record to work in settings where files are immutable, but still
provide a structured way of updating data stored inside.

Both base containers and patches are HDF5 files that are linked together
by some special attributes in the container root.
`IH5Record` is a class that wraps such a set of files. It features
* support for record creation and updating
* automatic handling of the patch mechanism (i.e., creating/finding corresponding files)
* transparent access to data in the record (possibly spanning multiple files)
"""
from __future__ import annotations

import json
import re
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Type, TypeVar, Union
from uuid import UUID, uuid1

import h5py
from pydantic import BaseModel, Field, PrivateAttr
from typing_extensions import Annotated, Final

from ..container.protocols import _OPEN_MODES, OpenMode
from ..schema.types import QualHashsumStr
from ..util.hashsums import qualified_hashsum
from .overlay import IH5Group, h5_copy_from_to

# the magic string we use to identify a valid container
FORMAT_MAGIC_STR: Final[str] = "ih5_v01"
"""Magic value at the beginning of the file to detect that an HDF5 file is valid IH5."""

USER_BLOCK_SIZE: Final[int] = 1024
"""Space to reserve at beginning of each HDF5 file in bytes.
Must be a power of 2 and at least 512 (required by HDF5)."""

T = TypeVar("T", bound="IH5Record")


def hashsum_file(filename: Path, skip_bytes: int = 0) -> str:
    """Compute hashsum of HDF5 file (ignoring the first `skip_bytes`)."""
    with open(filename, "rb") as f:
        f.seek(skip_bytes)
        return qualified_hashsum(f)


class IH5UserBlock(BaseModel):
    """IH5 metadata object parser and writer for the HDF5 user block.

    The user block stores technical administrative information linking together
    multiple HDF5 files that form a base container + patch sequence.

    The userblock begins with the magic format string, followed by a newline,
    followed by a string encoding the length of the user block, followed by a newline,
    followed by a serialized JSON object without newlines containing the metadata,
    followed by a NUL byte.
    """

    _userblock_size: int = PrivateAttr(default=USER_BLOCK_SIZE)
    """User block size claimed in the block itself (second line)."""

    record_uuid: UUID
    """UUID linking together multiple HDF5 files that form a (patched) record."""

    patch_index: Annotated[int, Field(ge=0)]
    """Index with the current revision number, i.e. the file is the n-th patch."""

    patch_uuid: UUID
    """UUID representing a certain state of the data in the record."""

    prev_patch: Optional[UUID]
    """UUID of the previous patch UUID (unless it is a base container, i.e. first one)."""

    hdf5_hashsum: Optional[QualHashsumStr] = None
    """Hashsum to verity integrity of the HDF5 data after the user block."""

    ub_exts: Dict[str, Any]
    """Any extra metadata to be stored in the user block, unvalidated in dicts.

    Subclasses must ensure that desired extra metadata is stored and loaded correctly.

    NOTE: In a merge of userblocks only newer extension section will be preserved!
    """

    @classmethod
    def create(cls, prev: Optional[IH5UserBlock] = None) -> IH5UserBlock:
        """Create a new user block for a base or patch container.

        If `prev` is None, will return a new base container block.
        Otherwise, will return a block linking back to the passed `prev` block.
        """
        ret = cls(
            patch_uuid=uuid1(),
            record_uuid=uuid1() if prev is None else prev.record_uuid,
            patch_index=0 if prev is None else prev.patch_index + 1,
            prev_patch=None if prev is None else prev.patch_uuid,
            ub_exts={},
        )
        return ret

    @classmethod
    def _read_head_raw(cls, stream, ub_size: int) -> Optional[Tuple[int, str]]:
        """Try reading user block.

        Args:
            stream: the open binary file stream
            ub_size: number of bytes to read

        Returns:
            (user block size claimed in block, embedded data until first NUL byte)
            or None, if block does not look right.
        """
        stream.seek(0)
        probe = stream.read(ub_size)
        dat = probe.decode("utf-8").split("\n")
        if len(dat) != 3 or dat[0] != FORMAT_MAGIC_STR:
            return None
        return (int(dat[1]), dat[2][: dat[2].find("\x00")])  # read until first NUL byte

    @classmethod
    def load(cls, filename: Path) -> IH5UserBlock:
        """Load a user block of the given HDF5 file."""
        with open(filename, "rb") as f:
            # try smallest valid UB size first
            head = cls._read_head_raw(f, 512)
            if head is None:
                raise ValueError(f"{filename}: it doesn't look like a valid IH5 file!")
            if head[0] > 512:  # if stored user block size is bigger, re-read
                head = cls._read_head_raw(f, head[0])
                assert head is not None
        ret = IH5UserBlock.parse_obj(json.loads(head[1]))
        ret._userblock_size = head[0]
        return ret

    def save(self, filename: Union[Path, str]):
        """Save this object in the user block of the given HDF5 file."""
        filename = Path(filename)
        dat_str = f"{FORMAT_MAGIC_STR}\n{self._userblock_size}\n{self.json()}"
        data = dat_str.encode("utf-8")
        assert len(data) < USER_BLOCK_SIZE
        with open(filename, "r+b") as f:
            # check that this HDF file actually has a user block
            check = f.read(4)
            if check == b"\x89HDF":
                raise ValueError(f"{filename}: no user block reserved, can't write!")
            f.seek(0)
            f.write(data)
            f.write(b"\x00")  # mark end of the data


class IH5Record(IH5Group):
    """Class representing a record, which consists of a collection of immutable files.

    One file is a base container (with no linked predecessor state),
    the remaining files are a linear sequence of patch containers.

    Runtime invariants to be upheld before/after each method call (after __init__):

    * all files of an instance are open for reading (until `close()` is called)
    * all files in `__files__` are in patch index order
    * at most one file is open in writable mode (if any, it is the last one)
    * modifications are possible only after `create_patch` was called
        and until `commit_patch` or `discard_patch` was called, and at no other time
    """

    # Characters that may appear in a record name.
    # (to be put into regex [..] symbol braces)
    _ALLOWED_NAME_CHARS = r"A-Za-z0-9\-"

    # filenames for a record named NAME are of the shape:
    # NAME[<PATCH_INFIX>.*]?<FILE_EXT>
    # NOTE: the first symbol of these must be one NOT in ALLOWED_NAME_CHARS!
    # This constraint is needed for correctly filtering filenames
    _PATCH_INFIX = ".p"
    _FILE_EXT = ".ih5"

    # core "wrapped" objects
    __files__: List[h5py.File]

    # attributes
    _closed: bool  # True after close()
    _allow_patching: bool  # false iff opened with "r"
    _ublocks: Dict[Path, IH5UserBlock]  # in-memory copy of HDF5 user blocks

    def __new__(cls, *args, **kwargs):
        ret = super().__new__(cls)
        ret._allow_patching = True
        ret.__files__ = []
        return ret

    def __eq__(self, o) -> bool:
        if not isinstance(o, IH5Group):
            return False
        return self._record._files == o._record._files

    @property
    def _has_writable(self):
        """Return True iff an uncommitted patch exists."""
        if not self.__files__:
            return False
        f = self.__files__[-1]
        return bool(f) and f.mode == "r+"

    @classmethod
    def _is_valid_record_name(cls, name: str) -> bool:
        """Return whether a record name is valid."""
        return re.match(f"^[{cls._ALLOWED_NAME_CHARS}]+$", name) is not None

    @classmethod
    def _base_filename(cls, record_path: Path) -> Path:
        """Given a record path, return path to canonical base container name."""
        return Path(f"{record_path}{cls._FILE_EXT}")

    @classmethod
    def _infer_name(cls, record_path: Path) -> str:
        return record_path.name.split(cls._FILE_EXT)[0].split(cls._PATCH_INFIX)[0]

    def _next_patch_filepath(self) -> Path:
        """Compute filepath for the next patch based on the previous one."""
        path = Path(self.__files__[0].filename)
        parent = path.parent
        patch_index = self._ublock(-1).patch_index + 1
        res = f"{parent}/{self._infer_name(path)}{self._PATCH_INFIX}{patch_index}{self._FILE_EXT}"
        return Path(res)

    def _ublock(self, obj: Union[h5py.File, int]) -> IH5UserBlock:
        """Return the parsed user block of a container file."""
        f: h5py.File = obj if isinstance(obj, h5py.File) else self.__files__[obj]
        return self._ublocks[Path(f.filename)]

    def _set_ublock(self, obj: Union[h5py.File, int], ub: IH5UserBlock):
        f: h5py.File = obj if isinstance(obj, h5py.File) else self.__files__[obj]
        self._ublocks[Path(f.filename)] = ub

    @classmethod
    def _new_container(cls, path: Path, ub: IH5UserBlock) -> h5py.File:
        """Initialize a fresh container file with reserved user block."""
        # create if does not exist, fail if it does
        f = h5py.File(path, mode="x", userblock_size=USER_BLOCK_SIZE)
        # close to pre-fill userblock
        f.close()
        ub.save(path)
        # reopen the container file
        return h5py.File(path, "r+")

    def _check_ublock(
        self,
        filename: Union[str, Path],
        ub: IH5UserBlock,
        prev: Optional[IH5UserBlock] = None,
        check_hashsum: bool = True,
    ):
        """Check given container file.

        If `prev` block is given, assumes that `ub` is from a patch container,
        otherwise from base container.
        """
        filename = Path(filename)
        # check presence+validity of record uuid (should be the same for all)
        if ub.record_uuid != self.ih5_uuid:
            msg = "'record_uuid' inconsistent! Mixed up records?"
            raise ValueError(f"{filename}: {msg}")

        # hash must match with HDF5 content (i.e. integrity check)
        if check_hashsum and ub.hdf5_hashsum is None:
            msg = "hdf5_checksum is missing!"
            raise ValueError(f"{filename}: {msg}")
        if ub.hdf5_hashsum is not None:
            chksum = hashsum_file(filename, skip_bytes=USER_BLOCK_SIZE)
            if ub.hdf5_hashsum != chksum:
                msg = "file has been modified, stored and computed checksum are different!"
                raise ValueError(f"{filename}: {msg}")

        # check patch chain structure
        if prev is not None:
            if ub.patch_index <= prev.patch_index:
                msg = "patch container must have greater index than predecessor!"
                raise ValueError(f"{filename}: {msg}")
            if ub.prev_patch is None:
                msg = "patch must have an attribute 'prev_patch'!"
                raise ValueError(f"{filename}: {msg}")
            # claimed predecessor uuid must match with the predecessor by index
            # (can compare as strings directly, as we checked those already)
            if ub.prev_patch != prev.patch_uuid:
                msg = f"patch for {ub.prev_patch}, but predecessor is {prev.patch_uuid}"
                raise ValueError(f"{filename}: {msg}")

    def _expect_open(self):
        if self._closed:
            raise ValueError("Record is not open!")

    def _clear(self):
        """Clear all contents of the record."""
        for k in self.attrs.keys():
            del self.attrs[k]
        for k in self.keys():
            del self[k]

    def _is_empty(self) -> bool:
        """Return whether this record currently contains any data."""
        return not self.attrs.keys() and not self.keys()

    @classmethod
    def _create(cls: Type[T], record: Union[Path, str], truncate: bool = False) -> T:
        """Create a new record consisting of a base container.

        The base container is exposed as the `writable` container.
        """
        record = Path(record)  # in case it was a str
        if not cls._is_valid_record_name(record.name):
            raise ValueError(f"Invalid record name: '{record.name}'")
        path = cls._base_filename(record)

        # if overwrite flag is set, check and remove old record if present
        if truncate and path.is_file():
            cls.delete_files(record)

        # create new container
        ret = cls.__new__(cls)
        super().__init__(ret, ret)
        ret._closed = False

        ub = IH5UserBlock.create(prev=None)
        ret._ublocks = {path: ub}

        ret.__files__ = [cls._new_container(path, ub)]
        return ret

    @classmethod
    def _open(cls: Type[T], paths: List[Path], **kwargs) -> T:
        """Open a record consisting of a base container + possible set of patches.

        Expects a set of full file paths forming a valid record.
        Will throw an exception in case of a detected inconsistency.

        Will open latest patch in writable mode if it lacks a hdf5 checksum.
        """
        if not paths:
            raise ValueError("Cannot open empty list of containers!")
        allow_baseless: bool = kwargs.pop("allow_baseless", False)

        ret = cls.__new__(cls)
        super().__init__(ret, ret)
        ret._closed = False

        ret._ublocks = {Path(path): IH5UserBlock.load(path) for path in paths}
        # files, sorted  by patch index order (important!)
        # if something is wrong with the indices, this will throw an exception.
        ret.__files__ = [h5py.File(path, "r") for path in paths]
        ret.__files__.sort(key=lambda f: ret._ublock(f).patch_index)
        # ----
        has_patches: bool = len(ret.__files__) > 1

        # check containers and relationship to each other:

        # check first container (it could be a base container and it has no predecessor)
        if not allow_baseless and ret._ublock(0).prev_patch is not None:
            msg = "base container must not have attribute 'prev_patch'!"
            raise ValueError(f"{ret.__files__[0].filename}: {msg}")
        ret._check_ublock(ret.__files__[0].filename, ret._ublock(0), None, has_patches)

        # check patches except last one (with checking the hashsum)
        for i in range(1, len(ret.__files__) - 1):
            filename = ret.__files__[i].filename
            ret._check_ublock(filename, ret._ublock(i), ret._ublock(i - 1), True)
        if has_patches:  # check latest patch (without checking hashsum)
            ret._check_ublock(
                ret.__files__[-1].filename, ret._ublock(-1), ret._ublock(-2), False
            )

        # now check whether the last container (patch or base or whatever) has a checksum
        if ret._ublock(-1).hdf5_hashsum is None:
            if kwargs.pop("reopen_incomplete_patch", False):
                # if opening in writable mode, allow to complete the patch
                f = ret.__files__[-1]
                path = ret.__files__[-1].filename
                f.close()
                ret.__files__[-1] = h5py.File(Path(path), "r+")

        # additional sanity check: container uuids must be all distinct
        cn_uuids = {ret._ublock(f).patch_uuid for f in ret.__files__}
        if len(cn_uuids) != len(ret.__files__):
            raise ValueError("Some patch_uuid is not unique, invalid file set!")
        # all looks good
        return ret

    # ---- public attributes and interface ----

    @property
    def ih5_uuid(self) -> UUID:
        """Return the common record UUID of the set of containers."""
        return self._ublock(0).record_uuid

    @property
    def ih5_files(self) -> List[Path]:
        """List of container filenames this record consists of."""
        return [Path(f.filename) for f in self.__files__]

    @property
    def ih5_meta(self) -> List[IH5UserBlock]:
        """Return user block metadata, in container patch order."""
        return [self._ublock(i).copy() for i in range(len(self.__files__))]

    @classmethod
    def find_files(cls, record: Path) -> List[Path]:
        """Return file names that look like they belong to the same record.

        This operation is based on purely syntactic pattern matching on file names
        that follow the default naming convention.

        Given a path `/foo/bar`, will find all containers in directory
        `/foo` whose name starts with `bar` followed by the correct file extension(s),
        such as `/foo/bar.ih5` and `/foo/bar.p1.ih5`.
        """
        record = Path(record)  # in case it was a str
        if not cls._is_valid_record_name(record.name):
            raise ValueError(f"Invalid record name: '{record.name}'")

        globstr = f"{record.name}*{cls._FILE_EXT}"  # rough wildcard pattern
        # filter out possible false positives (i.e. foobar* matching foo* as well)
        return [
            p
            for p in record.parent.glob(globstr)
            if re.match(f"^{record.name}[^{cls._ALLOWED_NAME_CHARS}]", p.name)
        ]

    @classmethod
    def list_records(cls, dir: Path) -> List[Path]:
        """Return paths of records found in the given directory.

        Will NOT recurse into subdirectories.

        This operation is based on purely syntactic pattern matching on file names
        that follow the default naming convention (i.e. just as `find_files`).

        Returned paths can be used as-is for opening the (supposed) record.
        """
        dir = Path(dir)  # in case it was a str
        if not dir.is_dir():
            raise ValueError(f"'{dir}' is not a directory")

        ret = []
        namepat = f"[{cls._ALLOWED_NAME_CHARS}]+(?=[^{cls._ALLOWED_NAME_CHARS}])"
        for p in dir.glob(f"*{cls._FILE_EXT}"):
            if m := re.match(namepat, p.name):
                ret.append(m[0])
        return list(map(lambda name: dir / name, set(ret)))

    def __init__(
        self, record: Union[str, Path, List[Path]], mode: OpenMode = "r", **kwargs
    ):
        """Open or create a record.

        This method uses `find_files` to infer the correct set of files syntactically.

        The open mode semantics are the same as for h5py.File.

        If the mode is 'r', then creating, committing or discarding patches is disabled.

        If the mode is 'a' or 'r+', then a new patch will be created in case the latest
        patch has already been committed.
        """
        super().__init__(self)

        if isinstance(record, list):
            if mode[0] == "w" or mode == "x":
                raise ValueError("Pass a prefix path for creating or overwriting!")
            paths = record
        else:
            paths = None
            path: Path = Path(record)

        if mode not in _OPEN_MODES:
            raise ValueError(f"Unknown file open mode: {mode}")

        if mode[0] == "w" or mode == "x":
            # create new or overwrite to get new
            ret = self._create(path, truncate=(mode == "w"))
            self.__dict__.update(ret.__dict__)
            return

        if mode == "a" or mode[0] == "r":
            if not paths:  # user passed a path prefix -> find files
                paths = self.find_files(path)  # type: ignore

            if not paths:  # no files were found
                if mode != "a":  # r/r+ need existing containers
                    raise FileNotFoundError(f"No files found for record: {path}")
                else:  # 'a' means create new if not existing (will be writable)
                    ret = self._create(path, truncate=False)
                    self.__dict__.update(ret.__dict__)
                    return

            # open existing (will be ro if everything is fine, writable if latest patch was uncommitted)
            want_rw = mode != "r"
            ret = self._open(paths, reopen_incomplete_patch=want_rw, **kwargs)
            self.__dict__.update(ret.__dict__)
            self._allow_patching = want_rw

            if want_rw and not self._has_writable:
                # latest patch was completed correctly -> make writable by creating new patch
                self.create_patch()

    @property
    def mode(self) -> Literal["r", "r+"]:
        return "r+" if self._allow_patching else "r"

    def close(self, commit: bool = True) -> None:
        """Close all files that belong to this record.

        If there exists an uncommited patch, it will be committed
        (unless `commit` is set to false).

        After this, the object may not be used anymore.
        """
        if self._closed:
            return  # nothing to do

        if self._has_writable and commit:
            self.commit_patch()
        for f in self.__files__:
            f.close()
        self.__files__ = []
        self._closed = True

    def _expect_not_ro(self):
        if self.mode == "r":
            raise ValueError("The container is opened as read-only!")

    def create_patch(self) -> None:
        """Create a new patch in order to update the record."""
        self._expect_open()
        self._expect_not_ro()
        if self._has_writable:
            raise ValueError("There already exists a writable container, commit first!")

        path = self._next_patch_filepath()
        ub = IH5UserBlock.create(prev=self._ublock(-1))
        self.__files__.append(self._new_container(path, ub))
        self._ublocks[path] = ub

    def _delete_latest_container(self) -> None:
        """Discard the current writable container (patch or base)."""
        cfile = self.__files__.pop()
        fn = cfile.filename
        del self._ublocks[Path(fn)]
        cfile.close()
        Path(fn).unlink()

    def discard_patch(self) -> None:
        """Discard the current incomplete patch container."""
        self._expect_open()
        self._expect_not_ro()
        if not self._has_writable:
            raise ValueError("No patch to discard!")
        if len(self.__files__) == 1:
            raise ValueError("Cannot discard base container! Just delete the file!")
            # reason: the base container provides record_uuid,
            # destroying it makes this object inconsistent / breaks invariants
            # so if this is done, it should not be used anymore.
        return self._delete_latest_container()

    def commit_patch(self, **kwargs) -> None:
        """Complete the current writable container (base or patch) for the record.

        Will perform checks on the new container and throw an exception on failure.

        After committing the patch is completed and cannot be edited anymore, so
        any further modifications must go into a new patch.
        """
        if kwargs:
            raise ValueError(f"Unknown keyword arguments: {kwargs}")

        self._expect_open()
        self._expect_not_ro()
        if not self._has_writable:
            raise ValueError("No patch to commit!")
        cfile = self.__files__[-1]
        filepath = Path(cfile.filename)
        cfile.close()  # must close it now, as we will write outside of HDF5 next

        # compute checksum, write user block
        chksum = hashsum_file(filepath, skip_bytes=USER_BLOCK_SIZE)
        self._ublocks[filepath].hdf5_hashsum = QualHashsumStr(chksum)
        self._ublocks[filepath].save(filepath)

        # reopen the container file now as read-only
        self.__files__[-1] = h5py.File(filepath, "r")

    def _fixes_after_merge(self, merged_file, ub):
        """Run hook for subclasses into merge process.

        The method is called after creating the merged container, but before
        updating its user block on disk.

        The passed userblock is a prepared userblock with updated HDF5 hashsum for the
        merged container and adapted prev_patch field, as will it be written to the file.
        Additional changes done to it in-place will be included.

        The passed filename can be used to perform additional necessary actions.
        """

    def merge_files(self, target: Path) -> Path:
        """Given a path with a record name, merge current record into new container.

        Returns new resulting container.
        """
        self._expect_open()
        if self._has_writable:
            raise ValueError("Cannot merge, please commit or discard your changes!")

        with type(self)(target, "x") as ds:
            source_node = self["/"]
            target_node = ds["/"]
            for k, v in source_node.attrs.items():  # copy root attributes
                target_node.attrs[k] = v
            for name in source_node.keys():  # copy each entity (will recurse)
                h5_copy_from_to(source_node[name], target_node, name)

            cfile = ds.ih5_files[0]  # store filename to override userblock afterwards

        # compute new merged userblock
        ub = self._ublock(-1).copy(update={"prev_patch": self._ublock(0).prev_patch})
        # update hashsum with saved new merged hdf5 payload
        chksum = hashsum_file(cfile, skip_bytes=USER_BLOCK_SIZE)
        ub.hdf5_hashsum = QualHashsumStr(chksum)

        self._fixes_after_merge(cfile, ub)  # for subclass hooks

        self._set_ublock(-1, ub)
        ub.save(cfile)
        return cfile

    @classmethod
    def delete_files(cls, record: Path):
        """Irreversibly(!) delete all containers matching the record path.

        This object is invalid after this operation.
        """
        for file in cls.find_files(record):
            file.unlink()

    def __repr__(self):
        return f"<IH5 record (mode {self.mode}) {self.__files__}>"

    # ---- context manager support (i.e. to use `with`) ----

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        # this will ensure that commit_patch() is called and the files are closed
        self.close()



"""Functionality to make widgets and dashboard work stand-alone in Jupyter."""

import wrapt

from ...container import MetadorContainer
from . import standalone


class Previewable(wrapt.ObjectProxy):
    """Wrapper to be used around MetadorContainer inside Jupyter.

    Will ensure that widgets can work in the notebook.
    """

    def __init__(self, container: MetadorContainer):
        super().__init__(container)

        if not standalone.running():
            standalone.run()

        if provider := standalone.container_provider():
            provider[str(self.metador.container_uuid)] = self

    def close(self, *args, **kwargs):
        if provider := standalone.container_provider():
            del provider[str(self.metador.container_uuid)]

        self.__wrapped__.close(*args, **kwargs)


__all__ = ["Previewable"]



"""Processing of entry points for Metador plugins."""
from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

from importlib_metadata import Distribution, entry_points

from ..schema.plugins import PluginPkgMeta, SemVerTuple
from .types import (
    EPGroupName,
    from_ep_group_name,
    is_metador_ep_group,
    to_ep_group_name,
)

_eps = entry_points()
"""All entry points."""

pkg_meta = {}
"""Collected infos about packages that provide plugins (filled by get_group)."""


def get_group(group_name: str) -> Dict[str, Any]:
    """Get a dict of all available entrypoints for a Metador plugin group."""
    ep_grp = to_ep_group_name(group_name)
    plugins: Dict[str, Any] = {}

    for ep in _eps.select(group=ep_grp):
        if ep.name in plugins:
            # TODO: will importlib_metadata even return colliding packages?
            # should be figured out (quite important to know)
            msg = f"{group_name}: a plugin named '{ep.name}' is already registered!"
            raise TypeError(msg)

        plugins[ep.name] = ep
        if ep.dist.name not in pkg_meta:
            pkg_meta[ep.dist.name] = PluginPkgMeta.for_package(ep.dist.name)

    return plugins


# ----


@dataclass
class DistMeta:
    name: str
    version: Tuple[int, int, int]
    plugins: Dict[str, List[str]]
    repository_url: Optional[str]


def distmeta_for(dist: Distribution) -> DistMeta:
    """Extract required metadata from importlib_metadata distribution object."""
    ver = dist.version
    if not re.fullmatch("[0-9]+\\.[0-9]+\\.[0-9]+", ver):
        msg = f"Invalid version string of {dist.name}: {ver}"
        raise TypeError(msg)
    parsed_ver: SemVerTuple = tuple(map(int, ver.split(".")))  # type: ignore

    # parse entry point groups
    epgs = filter(
        is_metador_ep_group,
        dist.entry_points.groups,
    )
    eps = {
        from_ep_group_name(EPGroupName(epg)): list(
            map(lambda x: x.name, dist.entry_points.select(group=epg))
        )
        for epg in epgs
    }

    repo_url: Optional[str] = None
    try:
        urls = dist.metadata.get_all("Project-URL")
        url = next(filter(lambda u: u.startswith("Repository,"), urls or []))
        repo_url = url.split()[1]  # from "Repository, http://..."
    except StopIteration:
        pass
    return DistMeta(
        name=dist.name, version=parsed_ver, plugins=eps, repository_url=repo_url
    )



from typing import Any, Dict, Literal, Optional

from pydantic.fields import ModelField

from ..util import is_public_name
from ..util.models import field_parent_type
from ..util.typing import get_annotations, is_enum, is_literal, is_subtype, unoptional
from .core import MetadataSchema


def _expect_schema_class(mcls):
    if not issubclass(mcls, MetadataSchema):
        raise TypeError("This decorator is for MetadataSchema subclasses!")


def _check_names_public(names):
    if priv_overrides := set(filter(lambda x: not is_public_name(x), names)):
        raise ValueError(f"Illegal private overrides: {priv_overrides}")


def make_mandatory(*names: str):
    """Make a field inherited from a base class mandatory if it is optional.

    The field must exist in a base class and must not be defined in the
    decorated class.

    Use this decorator instead of manually declaring an annotation,
    if all you need to do is making an existing field mandatory.
    """
    # NOTE: idea: could take a dict, then values are the new default for non-optional
    # but is this good? defaults are implicit optionality -> we discourage it, so no.
    _check_names_public(names)

    def make_fields_mandatory(mcls):
        _expect_schema_class(mcls)

        for name in names:
            if name not in mcls.__fields__:
                raise ValueError(f"{mcls.__name__} has no field named '{name}'!")
            if name in get_annotations(mcls):
                raise ValueError(
                    f"{mcls.__name__} already defines '{name}', cannot use decorator!"
                )

            hint = unoptional(field_parent_type(mcls, name))
            # update model and type hint (important for type analysis)
            mcls.__fields__[name].required = True
            mcls.__annotations__[name] = hint

        return mcls

    return make_fields_mandatory


def add_const_fields(consts: Dict[str, Any], *, override: bool = False):
    """Add constant fields to pydantic models.

    Must be passed a dict of field names and the constant values (only JSON-like types).

    Constant fields are optional during input.
    If present during parsing, they are be ignored and overriden with the constant.
    Constant fields are included in serialization, unless `exclude_defaults` is set.

    This can be used e.g. to attach JSON-LD annotations to schemas.

    Constant fields are inherited and may only be overridden by other constant fields
    using this decorator, they cannot become normal fields again.
    """
    _check_names_public(consts.keys())

    def add_fields(mcls):
        _expect_schema_class(mcls)

        # hacking it in-place approach:
        overridden = set()
        for name, value in consts.items():
            if field_def := mcls.__fields__.get(name):  # overriding a field
                # we allow to silently override of enum/literal types with suitable values
                # to support a schema design pattern of marked subclasses
                # but check that it is actually used correctly.
                enum_specialization = is_enum(field_def.type_)
                literal_specialization = is_literal(field_def.type_)

                valid_specialization = False
                if enum_specialization:
                    valid_specialization = isinstance(value, field_def.type_)
                elif literal_specialization:
                    lit_const = Literal[value]  # type: ignore
                    valid_specialization = is_subtype(lit_const, field_def.type_)

                if (
                    enum_specialization or literal_specialization
                ) and not valid_specialization:
                    msg = f"{mcls.__name__}.{name} cannot be overriden with '{value}', "
                    msg += f"because it is not a valid value of {field_def.type_}!"
                    raise TypeError(msg)

                # reject if not force override or allowed special cases
                if not (override or enum_specialization or literal_specialization):
                    msg = f"{mcls.__name__} already has a field '{name}'!"
                    msg += f" (override={override})"
                    raise ValueError(msg)

                else:  # new field
                    overridden.add(name)

            # this would force the exact constant on load
            # but this breaks parent compatibility if consts overridden!
            # ----
            # val = value.default if isinstance(value, FieldInfo) else value
            # ctype = Optional[make_literal(val)]  # type: ignore
            # ----
            # we simply ignore the constants as opaque somethings
            ctype = Optional[Any]  # type: ignore

            # configure pydantic field
            field = ModelField.infer(
                name=name,
                value=value,
                annotation=ctype,
                class_validators=None,
                config=mcls.__config__,
            )
            mcls.__fields__[name] = field
            # add type hint (important for our field analysis!)
            mcls.__annotations__[name] = field.type_
        ret = mcls

        # dynamic subclass approach:
        # ret = create_model(mcls.__name__, __base__=mcls, __module__=mcls.__module__, **consts)
        # if hasattr(mcls, "Plugin"):
        #     ret.Plugin = mcls.Plugin

        # to later distinguish "const" fields from normal fields:
        ret.__constants__.update(consts)
        return ret

    return add_fields


def override(*names: str):
    """Declare fields that are overridden (and not valid as subtypes).

    These are checked during plugin loading, in order to catch accidental
    overridden fields in schemas.
    """
    _check_names_public(names)

    def add_overrides(mcls):
        _expect_schema_class(mcls)
        mcls.__overrides__.update(set(names))
        return mcls

    return add_overrides



"""Common generic widgets.

These basically integrate many default widgets provided by panel/bokeh into Metador.
"""

import json
from typing import List, Set, Type

import panel as pn
from overrides import overrides
from panel.viewable import Viewable

from metador_core.schema import MetadataSchema

from ..plugins import schemas
from . import Widget

FileMeta = schemas.get("core.file", (0, 1, 0))
ImageFileMeta = schemas.get("core.imagefile", (0, 1, 0))


class FileWidget(Widget):
    """Simple widget based on (a subschema of) 'core.file'.

    Allows to state supported MIME types with less boilerplate.
    """

    class Plugin:
        # name and version must be overridden in subclasses
        supports = [FileMeta.Plugin.ref()]

    MIME_TYPES: Set[str] = set()
    """If non-empty, metadata objects must have a MIME type from this set."""

    FILE_EXTS: Set[str] = set()
    """If non-empty, filename must have an extension from this set."""

    @property
    def title(self) -> str:
        return self._meta.name or self._meta.filename or self._node.name

    @classmethod
    @overrides
    def supports_meta(cls, obj: MetadataSchema) -> bool:
        supported_mime = True
        if cls.MIME_TYPES:
            supported_mime = obj.encodingFormat in cls.MIME_TYPES
        supported_ext = False
        if cls.FILE_EXTS:
            supported_ext = obj.filename.endswith(tuple(cls.FILE_EXTS))
        # either supported mime or supported ext is enough
        return supported_mime or supported_ext


# wrap simple generic widgets from panel:

# pass content:


class MarkdownWidget(FileWidget):
    class Plugin(FileWidget.Plugin):
        name = "core.file.text.md"
        version = (0, 1, 0)

    MIME_TYPES = {"text/markdown", "text/x-markdown"}
    FILE_EXTS = {".md", ".markdown"}

    @overrides
    def show(self) -> Viewable:
        return pn.pane.Markdown(
            self.file_data().decode("utf-8"), max_width=self._w, max_height=self._h
        )


class HTMLWidget(FileWidget):
    class Plugin(FileWidget.Plugin):
        name = "core.file.text.html"
        version = (0, 1, 0)

    MIME_TYPES = {"text/html"}
    FILE_EXTS = {".htm", ".html"}

    @overrides
    def show(self) -> Viewable:
        return pn.pane.HTML(
            self.file_data().decode("utf-8"), max_width=self._w, max_height=self._h
        )


class CodeWidget(FileWidget):
    class Plugin(FileWidget.Plugin):
        name = "core.file.text.code"
        version = (0, 1, 0)

    @classmethod
    @overrides
    def supports_meta(cls, obj: MetadataSchema) -> bool:
        return obj.encodingFormat.startswith("text")

    def show(self) -> Viewable:
        return pn.widgets.Ace(
            filename=self._meta.filename,
            readonly=True,
            value=self.file_data().decode("utf-8"),
            width=self._w,
            height=self._h,
        )


class DispatcherWidget(Widget):
    """Meta-widget to dispatch a node+metadata object to a more specific widget.

    Make sure that the dispatcher widget is probed before the widgets it can
    dispatch to.

    This works if and only if for each widget in listed in `WIDGETS`:
    * the plugin name of the dispatcher is a prefix of the widget name, or
    * the widget has `primary = False` and thus is not considered by the dashboard.
    """

    WIDGETS: List[Type[Widget]]
    """Widgets in the order they should be tested."""

    def dispatch(self, w_cls: Type[Widget]) -> Widget:
        """Dispatch to another widget (used by meta-widgets)."""
        return w_cls(
            self._node,
            "",
            server=self._server,
            metadata=self._meta,
            max_width=self._w,
            max_height=self._h,
        )

    @classmethod
    @overrides
    def supports_meta(cls, obj: MetadataSchema) -> bool:
        return any(map(lambda w: w.supports_meta(obj), cls.WIDGETS))

    @overrides
    def setup(self):
        for w_cls in self.WIDGETS:
            if w_cls.supports_meta(self._meta):
                self._widget = self.dispatch(w_cls)
                break

    @overrides
    def show(self) -> Viewable:
        return self._widget.show()


class TextWidget(DispatcherWidget, FileWidget):
    class Plugin(FileWidget.Plugin):
        name = "core.file.text"
        version = (0, 1, 0)

    WIDGETS = [MarkdownWidget, HTMLWidget, CodeWidget]

    def show(self) -> Viewable:
        return super().show()


class JSONWidget(FileWidget):
    class Plugin(FileWidget.Plugin):
        name = "core.file.json"
        version = (0, 1, 0)

    MIME_TYPES = {"application/json"}

    @overrides
    def show(self) -> Viewable:
        return pn.pane.JSON(
            json.loads(self.file_data()),
            name=self.title,
            max_width=self._w,
            max_height=self._h,
            hover_preview=True,
            depth=-1,
        )


# pass URL:


class PDFWidget(FileWidget):
    class Plugin(FileWidget.Plugin):
        name = "core.file.pdf"
        version = (0, 1, 0)

    MIME_TYPES = {"application/pdf"}

    @overrides
    def show(self) -> Viewable:
        return pn.pane.PDF(self.file_url(), width=self._w, height=self._h)


class ImageWidget(FileWidget):
    class Plugin(FileWidget.Plugin):
        name = "core.file.image"
        version = (0, 1, 0)

    MIME_TYPES = {"image/jpeg", "image/png", "image/gif", "image/svg+xml"}
    PANEL_WIDGET = {
        "image/jpeg": pn.pane.JPG,
        "image/png": pn.pane.PNG,
        "image/gif": pn.pane.GIF,
        "image/svg+xml": pn.pane.SVG,
    }

    @overrides
    def show(self) -> Viewable:
        return self.PANEL_WIDGET[self._meta.encodingFormat](
            self.file_url(),
            width=self._w,
            height=self._h,
            alt_text="Sorry. Something went wrong while loading the resource",
        )


# NOTE: Panel based widget does not work.
# see - https://github.com/holoviz/panel/issues/3203
# using HTML based audio & video panes for the audio and video widgets respectively
class AudioWidget(FileWidget):
    class Plugin(FileWidget.Plugin):
        name = "core.file.audio"
        version = (0, 1, 0)

    MIME_TYPES = {"audio/mpeg", "audio/ogg", "audio/webm"}

    @overrides
    def show(self) -> Viewable:
        return pn.pane.HTML(
            f"""
                <audio controls>
                    <source src={self.file_url()} type={self._meta.encodingFormat}>
                </audio>
            """,
        )


class VideoWidget(FileWidget):
    class Plugin(FileWidget.Plugin):
        name = "core.file.video"
        version = (0, 1, 0)

    MIME_TYPES = {"video/mp4", "video/ogg", "video/webm"}

    @overrides
    def show(self) -> Viewable:
        return pn.pane.HTML(
            f"""
                <video width={self._w} height={self._h} controls style="object-position: center; object-fit:cover;">
                <source src={self.file_url()} type={self._meta.encodingFormat}>
                </video>
            """,
        )



"""Plugin group for metadata harvesters."""
from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Set,
    Type,
    TypeVar,
    Union,
)

from overrides import overrides
from pydantic import Extra, ValidationError
from typing_extensions import TypeAlias

from ..plugin import interface as pg
from ..plugins import schemas
from ..schema import MetadataSchema
from ..schema.core import BaseModelPlus
from ..schema.partial import PartialFactory, PartialModel
from ..schema.plugins import PluginRef

HARVESTER_GROUP_NAME = "harvester"


if TYPE_CHECKING:
    SchemaPluginRef: TypeAlias = PluginRef
else:
    SchemaPluginRef = schemas.PluginRef


class HarvesterPlugin(pg.PluginBase):
    returns: SchemaPluginRef
    """Schema returned by this harvester."""


S = TypeVar("S", bound=MetadataSchema)


class HarvesterArgs(BaseModelPlus):
    """Base class for harvester arguments."""

    class Config(BaseModelPlus.Config):
        extra = Extra.forbid


class HarvesterArgsPartial(PartialFactory):
    """Base class for partial harvester arguments."""

    base_model = HarvesterArgs


class HarvesterMetaMixin(type):
    _args_partials: Dict[Type, Type] = {}

    @property
    # @cache not needed, partials take care of that themselves
    def PartialArgs(self):
        """Access the partial schema based on the current schema."""
        return HarvesterArgsPartial.get_partial(self.Args)


class HarvesterMeta(HarvesterMetaMixin, ABCMeta):
    ...


class Harvester(ABC, metaclass=HarvesterMeta):
    """Base class for metadata harvesters.

    A harvester is a class that can be instantiated to extract
    metadata according to some schema plugin (the returned schema
    must be defined as the `Plugin.returns` attribute)

    Override the inner `Harvester.Args` class with a subclass to add arguments.
    This works exactly like schema definition and is simply
    another pydantic model, so you can use both field and root
    validators to check whether the arguments are making sense.

    If all you need is getting a file path, consider using
    `FileHarvester`, which already defines a `filepath` argument.

    Override the `run()` method to implement the metadata harvesting
    based on a validated configuration located at `self.args`.
    """

    Plugin: ClassVar[HarvesterPlugin]

    Args: TypeAlias = HarvesterArgs
    """Arguments to be passed to the harvester."""

    args: Union[Harvester.Args, HarvesterArgsPartial]

    @property
    def schema(self):
        """Partial schema class returned by this harvester.

        Provided for implementation convenience.
        """
        return schemas[self.Plugin.returns.name].Partial

    # ----
    # configuring and checking harvester instances:

    def __init__(self, **kwargs):
        """Initialize harvester with (partial) configuration."""
        self.args = type(self).PartialArgs.parse_obj(kwargs)

    def __call__(self, **kwargs):
        """Return copy of harvester with updated configuration."""
        args = type(self).PartialArgs.parse_obj(kwargs)
        merged = self.args.merge_with(args)
        return type(self)(**merged.dict())

    def __repr__(self):
        return f"{type(self).__name__}(args={repr(self.args)})"

    def harvest(self):
        """Check provided arguments and run the harvester.

        Call this when a harvester is configured and it should be executed.
        """
        try:
            self.args = self.args.from_partial()
        except ValidationError as e:
            hname = type(self).__name__
            msg = f"{hname} configuration incomplete or invalid:\n{str(e)}"
            raise ValueError(msg)
        return self.run()

    # ----
    # to be overridden

    @abstractmethod
    def run(self) -> MetadataSchema:
        """Do the harvesting according to instance configuration and return metadata.

        Override this method with your custom metadata harvesting logic, based
        on configuration provided in `self.args`.

        All actual, possibly time-intensive harvesting computations (accessing
        resources, running external code, etc.) MUST be performed in this method.

        Ensure that your harvester does not interfere with anything else, such
        as other harvesters running on possibly the same data - open resources
        such as files in read-only mode. You SHOULD avoid creating tempfiles, or
        ensure that these are managed, fresh and cleaned up when you are done
        (e.g. if you need to run external processes that dump outputs to a
        file).

        Returns:
            A fresh instance of type `self.schema` containing harvested metadata.
        """


class FileHarvester(Harvester):
    """Harvester for processing a single file path.

    The file path is not provided or set during __init__,
    but instead is passed during harvest_file.
    """

    class Args(Harvester.Args):
        filepath: Path


class MetadataLoader(FileHarvester):
    _partial_schema: PartialModel
    _sidecar_func: Callable[[Path], Path]

    def run(self):
        path = self.args.filepath
        func = type(self).__dict__.get("_sidecar_func")
        used_path = func(path)
        if not used_path.is_file():
            return self._partial_schema()
        return self._partial_schema.parse_file(used_path)


# ----


def _schema_arg(obj: Union[str, Type[MetadataSchema]]) -> Type[MetadataSchema]:
    """Return installed schema by name or same schema as passed.

    Ensures that passed object is in fact a valid, installed schema.
    """
    name = obj if isinstance(obj, str) else obj.Plugin.name
    return schemas[name]


class PGHarvester(pg.PluginGroup[Harvester]):
    """Harvester plugin group interface."""

    class Plugin:
        name = HARVESTER_GROUP_NAME
        version = (0, 1, 0)
        requires = [PluginRef(group="plugingroup", name="schema", version=(0, 1, 0))]
        plugin_class = Harvester
        plugin_info_class = HarvesterPlugin

    def __post_init__(self):
        self._harvesters_for: Dict[str, Set[str]] = {}

    def plugin_deps(self, plugin):
        if p := plugin.Plugin.returns:
            return {p}

    @overrides
    def check_plugin(self, ep_name: str, plugin: Type[Harvester]):
        hv_ref = plugin.Plugin.returns

        schema_name = hv_ref.name
        schema = schemas[schema_name]
        if not schema:
            raise TypeError(f"{ep_name}: Schema '{schema_name}' not installed!")

        inst_ref = schema.Plugin.ref()
        if not inst_ref.supports(hv_ref):
            msg = f"{ep_name}: Installed schema {inst_ref} incompatible with harvester schema {hv_ref}!"
            raise TypeError(msg)

    @overrides
    def init_plugin(self, plugin):
        """Add harvester to harvester lookup table."""
        h_name = plugin.Plugin.name
        schema = plugin.Plugin.returns
        if schema.name not in self._harvesters_for:
            self._harvesters_for[schema.name] = set()
        self._harvesters_for[schema.name].add(h_name)

    def for_schema(
        self,
        schema: Union[str, MetadataSchema],
        *,
        include_children: bool = False,
        include_parents: bool = False,
    ) -> Set[str]:
        """List installed harvesters for the given metadata schema.

        To extend the query to parent or child schemas, set `include_children`
        and/or `include_parents` accordingly.

        Harvesters for child schemas are always compatible with the schema.
        (assuming correct implementation of the child schemas),

        Harvesters for parent schemas can be incompatible in some circumstances
        (specifically, if the parent accepts values for some field that are
        forbidden in the more specialized schema).

        Args:
            schema: schema name or class for which to return harvesters
            include_children: Also include results for installed children
            include_parents: Also include results for parent schemas

        Returns:
            Set of harvesters.
        """
        schema_name = _schema_arg(schema).Plugin.name
        ret = set(self._harvesters_for[schema_name])
        if include_children:
            for child in schemas.children(schema_name):
                ret = ret.union(self.for_schema(child))
        if include_parents:
            for parent in schemas.parent_path(schema_name)[:-1]:
                ret = ret.union(self.for_schema(parent))
        return ret


def _harvest_source(schema: Type[S], obj: Union[Path, Harvester]) -> S:
    """Harvest partial metadata from the passed Harvester or Path into target schema."""
    if not isinstance(obj, (Path, Harvester)):
        raise ValueError("Metadata source must be a Harvester or a Path!")
    if isinstance(obj, Path):
        obj = metadata_loader(schema)(filepath=obj)
    return obj.harvest()


def _identity(x):
    return x


# harvesting helpers:


def def_sidecar_func(path: Path):
    """Return sidecar filename for a given file path using the default convention.

    This means, the path gets a '_meta.yaml' suffix attached.
    """
    return Path(f"{str(path)}_meta.yaml")


def metadata_loader(
    schema: Type[MetadataSchema], *, use_sidecar: bool = False, sidecar_func=None
) -> Type[MetadataLoader]:
    """Return harvester for partial metadata files of specified schema.

    Will return an empty partial schema in case the file does not exist.

    The fields provided in the file must be valid, or an exception will be raised.
    This is to avoid the situation where a user intends to add or override
    harvester metadata, but the fields are silently ignored, leading to possible
    surprise and confusion.

    By default, the returned harvester will attempt to parse the provided file.

    Set `use_sidecar` if you intend to pass it filenames of data files, but
    want it to instead load metadata from a sidecar file.

    Provide a `sidecar_func` if you want to change the used sidecar file convention.
    By default, `def_sidecar_func` is used to transform the path.
    """
    used_sidecar_f = _identity
    if use_sidecar:
        used_sidecar_f = sidecar_func if sidecar_func else def_sidecar_func

    class XLoader(MetadataLoader):
        _partial_schema = schema.Partial
        _sidecar_func = used_sidecar_f

    name = f"{schema.__name__}Loader"
    XLoader.__qualname__ = XLoader.__name__ = name
    return XLoader


def configure(*harvesters: Union[Harvester, Type[Harvester]], **kwargs):
    """Given a sequence of harvesters, configure them all at once.

    Can be used to set the same parameter in all of them easily.
    """
    return (h(**kwargs) for h in harvesters)


def file_harvester_pipeline(
    *hvs: Union[FileHarvester, Type[FileHarvester]]
) -> Callable[[Path], Any]:
    """Generate a harvesting pipeline for a file.

    Args:
        hvs: FileHarvester classes or pre-configured instances to use.
            The passed objects must be preconfigured as needed,
            except for fixing a filepath (it will be overwritten).

    Returns:
        Function that takes a file path and will return
        the harvesters configured for the passed path.
    """
    return lambda path: configure(*hvs, filepath=path)


def harvest(
    schema: Type[S],
    sources: Iterable[Union[Path, Harvester]],
    *,
    ignore_invalid: bool = False,
    return_partial: bool = False,
) -> S:
    """Run a harvesting pipeline and return combined results.

    Will run the harvesters in the passed order, combining results.

    In general you can expect that if two harvesters provide the same field,
    the newer value by a later harvester will overwrite an existing one from an
    earlier harvester, or the values are combined in a suitable way.

    If converting some Harvester result to the desired schema fails,
    a `ValidationError` will be raised.  To change that behaviour,
    set `ignore_invalid`, will make sure that suitable fields are still used,
    even if the given object as a whole is not fully parsable.

    Note that `ignore_invalid` only affects the conversion AFTER running the
    harvesters, it does NOT affect the way the harvesters treat invalid metadata.

    By default, it is assumed that the result of the whole pipeline can be
    converted into a fully valid non-partial schema instance.
    If this final conversion fails, an exception will be raised.
    To prevent this conversion, set `return_partial`. In that case, you will
    get the partial instance as-is and can manually call `from_partial()`.

    Args:
        schema: Class of schema to be returned.
        sources: List of sources (Paths or Harvester instances).
        ignore_invalid: Whether to ignore invalid fields in Harvester outputs.
        return_partial: Whether to return raw partial instance (default: False).

    Returns:
        Metadata object with combined results.
    """

    def cast(meta):
        # configure cast to ignore or not ignore invalid fields
        return schema.Partial.cast(meta, ignore_invalid=ignore_invalid)

    # collect partial metadata (NOTE: in principle, this could be parallelized)
    results = map(lambda s: cast(_harvest_source(schema, s)), sources)

    # accumulate collected and casted results in provided order
    merged = schema.Partial.merge(*results)

    # retrieve (completed) metadata model
    return merged if return_partial else merged.from_partial()



"""Central place for convenient access to all registered plugin groups.

For example, to access the `schema` plugingroup, you can use:

`from metador_core.plugins import schemas`
"""
from typing import TYPE_CHECKING, Dict, List, TypeVar, cast

import wrapt

from .plugin.interface import PG_GROUP_NAME, AnyPluginRef, PluginGroup, plugin_args

S = TypeVar("S", bound=PluginGroup)


class PGPluginGroup(wrapt.ObjectProxy):
    """PluginGroup plugin group.

    This wrapper returns instances of other loaded plugin groups.

    In the esoteric case that you need to access the actual plugingroup class
    that gives out *classes* instead of instances (like all other plugingroups),
    request the "plugingroup" plugingroup. But usually you will not want this.
    """

    _self_groups: Dict[AnyPluginRef, PluginGroup]

    def __reset__(self):
        self._self_groups.clear()
        self.__init__()

    def __init__(self):
        # initialize the meta-plugingroup
        from .plugin.interface import _plugin_groups, create_pg

        create_pg(PluginGroup)
        pgpg_ref = AnyPluginRef(
            group=PG_GROUP_NAME,
            name=PluginGroup.Plugin.name,
            version=PluginGroup.Plugin.version,
        )

        # wire it up with this wrapper
        self._self_groups = _plugin_groups
        self.__wrapped__ = _plugin_groups[pgpg_ref]

    # ----
    def get(self, key, version=None):
        """Get a registered plugin group by name."""
        key_, vers = plugin_args(key, version)
        if key_ == self.name and (vers is None or vers == self.Plugin.version):
            return self
        try:
            if grp_cls := self.__wrapped__._get_unsafe(key_, vers):
                # now if the PG was not existing, it is + is stored in _self_groups
                return cast(S, self._self_groups.get(grp_cls.Plugin.ref()))
        except KeyError:
            return None

    def __getitem__(self, key) -> PluginGroup:
        # call wrapped '__getitem__' with this object to use its 'get'
        return PluginGroup.__getitem__(self, key)  # type: ignore

    def values(self):
        # same idea, this uses '__getitem__'
        return PluginGroup.values(self)

    def items(self):
        # same idea, this uses '__getitem__'
        return PluginGroup.items(self)

    def is_plugin(self, obj):
        return obj in self.values()


# access to available plugin groups:

plugingroups: PGPluginGroup = PGPluginGroup()

# help mypy (obviously only for groups in this package):
# NOTE: this would be better: https://github.com/python/mypy/issues/13643
if TYPE_CHECKING:  # pragma: no cover
    from .harvester import PGHarvester
    from .packer import PGPacker
    from .schema.pg import PGSchema
    from .widget import PGWidget

    schemas: PGSchema
    harvesters: PGHarvester
    widgets: PGWidget
    packers: PGPacker

# ----
# Now some magic to lift all other groups to module level,
# this allows to import like: from metador_core.plugins import schemas

# define what to import with *
__all__ = list(sorted(map(lambda ref: f"{ref.name}s", plugingroups.keys())))


def __dir__() -> List[str]:
    # show the existing plugin groups for tab completion in e.g. ipython
    return __all__


def __getattr__(key: str):
    # get desired plugin group and add it as module attribute
    # (i.e. this is called at most once per group)
    if not isinstance(key, str) or key[-1] != "s":
        raise AttributeError(key)
    if group := plugingroups.get(key[:-1]):
        globals()["__annotations__"][key] = type(group)
        globals()[key] = group
        return group
    raise AttributeError(key)



"""Schema.org-compatible common metadata schemas.

Supports a subset of commonly useful fields.

Adds almost no constraints beyond the spec, except for fixing a multiplicity for fields.

Intended to serve as the basis for more specific schemas.

Note that this schemas ARE NOT able to parse arbitrary schema.org-aligned metadata,
their purpose is to ensure that successfully parsed input is semantically enriched.

See schema.org official documentation for full explanation and list of all fields.
"""
from __future__ import annotations

from datetime import date, datetime, time
from typing import List, Optional, Set, Union

from pydantic import AnyHttpUrl, NonNegativeInt
from typing_extensions import TypeAlias

from ..ld import LDIdRef, LDOrRef, LDSchema, ld_decorator
from ..types import Bool, Duration, Float, Int, NonEmptyStr

CTX_URL_SCHEMAORG = "https://schema.org"

schemaorg = ld_decorator(context=CTX_URL_SCHEMAORG)

# ----

URL: TypeAlias = AnyHttpUrl
Text: TypeAlias = NonEmptyStr
Number: TypeAlias = Union[Int, Float]

DateOrDatetime = Union[date, datetime]
TimeOrDatetime = Union[time, datetime]

# ----


@schemaorg(type="Thing")
class Thing(LDSchema):
    """See https://schema.org/Thing for field documentation."""

    name: Optional[Text]
    """Name, title or caption of the entity."""

    identifier: Optional[Union[URL, Text]]  # can't put PropertyValue here, weird bug
    """Arbitrary identifier of the entity.

    Prefer @id if the identifier is web-resolvable, or use more
    specific fields if available."""

    url: Optional[URL]
    """URL of the entity."""

    description: Optional[Text]
    """Description of the entity."""

    # ----

    alternateName: Optional[List[Text]]
    """Known aliases of the entity."""

    sameAs: Optional[List[URL]]


class ValueCommon(Thing):
    """Common properties of multiple *Value classes.

    For some reason these have no common ancestor in schema.org.
    """

    value: Optional[Union[Bool, Number, Text, StructuredValue]]

    # valueReference: Optional[]

    minValue: Optional[Number]
    """Minimal value of property this value corresponds to."""

    maxValue: Optional[Number]
    """Maximal value of property this value corresponds to."""

    unitCode: Optional[Union[URL, Text]]
    """UN/CEFACT Common Code (3 characters) or URL.

    Other codes may be used with a prefix followed by a colon."""

    unitText: Optional[Text]
    """String indicating the unit of measurement.

    Useful if no standard unitCode can be provided.
    """


@schemaorg(type="StructuredValue")
class StructuredValue(ValueCommon):
    """See https://schema.org/StructuredValue for field documentation."""


@schemaorg(type="QuantitativeValue")
class QuantitativeValue(StructuredValue):
    """See https://schema.org/QuantitativeValue for field documentation."""


@schemaorg(type="PropertyValue")
class PropertyValue(StructuredValue):
    """Use 'name' for the property name and 'description' for alternative human-readable value.

    See https://schema.org/PropertyValue for field documentation.
    """

    propertyID: Optional[Union[URL, Text]]
    """A commonly used identifier for the characteristic represented by the property,
    e.g. a manufacturer or a standard code for a property."""

    measurementTechnique: Optional[Union[URL, Text]]
    """A technique or technology used in a Dataset (or DataDownload, DataCatalog),
    corresponding to the method used for measuring the corresponding variable(s)
    (described using variableMeasured).

    This is oriented towards scientific and scholarly dataset publication but
    may have broader applicability; it is not intended as a full representation
    of measurement, but rather as a high level summary for dataset discovery.
    """


# ----


@schemaorg(type="Organization")
class Organization(Thing):
    """See https://schema.org/Organization for field documentation."""

    address: Optional[Text]
    """Address of the organization."""


@schemaorg(type="Person")
class Person(Thing):
    """See https://schema.org/Person for field documentation."""

    givenName: Optional[Text]
    """Given name, typically the first name of a Person."""

    familyName: Optional[Text]
    """Family name of a Person."""

    additionalName: Optional[Text]
    """Additional name for a Person, e.g. for a middle name."""

    email: Optional[Text]
    """E-mail address."""

    affiliation: Optional[LDOrRef[Organization]]
    """An organization this person is affiliated with."""


OrgOrPerson = Union[Person, Organization]

# ----


@schemaorg(type="CreativeWork")
class CreativeWork(Thing):
    """See https://schema.org/CreativeWork for field documentation."""

    version: Optional[Union[NonNegativeInt, Text]]
    """Version of this work.

    Either an integer, or a version string, e.g. "1.0.5".

    When using version strings, follow https://semver.org
    whenever applicable.
    """

    citation: Optional[Set[Union[LDOrRef[CreativeWork], Text]]]
    """Citation or reference to another creative work, e.g.
    another publication, scholarly article, etc."""

    # search

    abstract: Optional[Text]
    """A short description that summarizes the creative work."""

    keywords: Optional[Set[Text]]
    """Keywords or tags to describe this creative work."""

    # people

    author: Optional[List[LDOrRef[OrgOrPerson]]]
    """People responsible for the work, e.g. in research,
    the people who would be authors on the relevant paper."""

    contributor: Optional[List[LDOrRef[OrgOrPerson]]]
    """Additional people who contributed to the work, e.g.
    in research, the people who would be in the acknowledgements
    section of the relevant paper."""

    maintainer: Optional[List[LDOrRef[OrgOrPerson]]]
    producer: Optional[List[LDOrRef[OrgOrPerson]]]
    provider: Optional[List[LDOrRef[OrgOrPerson]]]
    publisher: Optional[List[LDOrRef[OrgOrPerson]]]
    sponsor: Optional[List[LDOrRef[OrgOrPerson]]]
    editor: Optional[List[LDOrRef[Person]]]

    # date

    dateCreated: Optional[DateOrDatetime]
    dateModified: Optional[DateOrDatetime]
    datePublished: Optional[DateOrDatetime]

    # legal

    copyrightHolder: Optional[LDOrRef[OrgOrPerson]]
    copyrightYear: Optional[Int]
    copyrightNotice: Optional[Text]
    license: Optional[Union[URL, LDOrRef[CreativeWork]]]

    # provenance

    about: Optional[Set[LDOrRef[Thing]]]
    subjectOf: Optional[Set[LDOrRef[CreativeWork]]]
    hasPart: Optional[Set[LDOrRef[CreativeWork]]]
    isPartOf: Optional[Set[Union[URL, LDOrRef[CreativeWork]]]]
    isBasedOn: Optional[Set[Union[URL, LDOrRef[CreativeWork]]]]


# ----


@schemaorg(type="DefinedTermSet")
class DefinedTermSet(CreativeWork):
    """See https://schema.org/DefinedTermSet for field documentation."""

    hasDefinedTerm: List[LDOrRef[DefinedTerm]]


@schemaorg(type="DefinedTerm")
class DefinedTerm(Thing):
    """See https://schema.org/DefinedTerm for field documentation."""

    # NOTE: also use name and description

    termCode: Text
    """A code that identifies this DefinedTerm within a DefinedTermSet."""

    inDefinedTermSet: Optional[Union[URL, LDIdRef]]  # ref to a DefinedTermSet
    """A DefinedTermSet that contains this term."""


@schemaorg(type="CategoryCodeSet")
class CategoryCodeSet(DefinedTermSet):
    """See https://schema.org/CategoryCodeSet for field documentation."""

    hasCategoryCode: List[LDOrRef[CategoryCode]]


@schemaorg(type="CategoryCode")
class CategoryCode(DefinedTerm):
    """See https://schema.org/CategoryCode for field documentation."""

    codeValue: Text
    """A short textual code that uniquely identifies the value."""

    inCodeSet: Optional[Union[URL, LDIdRef]]  # ref to a CategoryCodeSet
    """A CategoryCodeSet that contains this category code."""


# ----


@schemaorg(type="MediaObject")
class MediaObject(CreativeWork):
    """See https://schema.org/MediaObject for field documentation."""

    contentSize: Optional[Int]
    """Size of the object in bytes."""

    sha256: Optional[Text]
    """Sha256 hashsum string of the object."""

    encodingFormat: Optional[Union[URL, Text]]
    """MIME type, or if the format is too niche or no standard MIME type is
    defined, an URL pointing to a description of the format."""

    width: Optional[QuantitativeValue]
    """Width of the entity."""

    height: Optional[QuantitativeValue]
    """Height of the entity."""

    bitrate: Optional[Text]
    """Bitrate of the entity (e.g. for audio or video)."""

    duration: Optional[Duration]
    """Duration of the entity (e.g. for audio or video)."""

    startTime: Optional[TimeOrDatetime]
    """Physical starting time, e.g. of a recording or measurement."""

    endTime: Optional[TimeOrDatetime]
    """Physical ending time, e.g. of a recording or measurement."""


@schemaorg(type="Dataset")
class Dataset(CreativeWork):
    """See https://schema.org/Dataset for field documentation."""

    distribution: Optional[URL]  # NOTE: for top level description could link to repo
    """Downloadable form of this dataset, at a specific location, in a specific format."""

    variableMeasured: Optional[List[Union[Text, PropertyValue]]]
    """Variables that are measured in the dataset."""


# ----


@schemaorg(type="Product")
class Product(Thing):
    """See https://schema.org/Product for field documentation."""

    productID: Optional[Text]
    """The product identifier, such as ISBN."""

    # properties

    category: Optional[Union[Text, URL, CategoryCode, Thing]]
    """A category for the item.

    Greater signs or slashes can be used to informally indicate a category hierarchy.
    """

    material: Optional[Union[URL, Text, Product]]
    """A material that something is made from, e.g. leather, wool, cotton, paper."""

    pattern: Optional[Union[DefinedTerm, Text]]
    """A pattern that something has, for example 'polka dot', 'striped', 'Canadian flag'.

    Values are typically expressed as text, although links to controlled value schemes are also supported.
    """

    width: Optional[QuantitativeValue]
    height: Optional[QuantitativeValue]
    depth: Optional[QuantitativeValue]

    weight: Optional[QuantitativeValue]
    color: Optional[Text]

    additionalProperty: Optional[List[PropertyValue]]
    """A property-value pair representing an additional characteristic of the entity, e.g. a product feature or another characteristic for which there is no matching property in schema.org."""

    # meta

    productionDate: Optional[DateOrDatetime]
    releaseDate: Optional[DateOrDatetime]

    isRelatedTo: Optional[LDOrRef[Product]]
    isSimilarTo: Optional[LDOrRef[Product]]



import pytest


def test_schema_pg(plugingroups_test):
    # BibMeta has DirMeta as parent. we use that for testing
    schemas = plugingroups_test.get("schema")
    BibMeta = schemas.get("core.bib")
    DirMeta = schemas.get("core.dir")
    b_ref = BibMeta.Plugin.ref()
    d_ref = DirMeta.Plugin.ref()

    # parentpath
    expected = [d_ref, b_ref]
    assert schemas.parent_path(b_ref.name, b_ref.version) == expected
    assert schemas.parent_path(b_ref) == expected
    assert schemas.parent_path(BibMeta) == expected

    assert schemas.parent_path(d_ref.name, d_ref.version) == [d_ref]

    with pytest.raises(KeyError):  # no such version of schema
        schemas.parent_path(b_ref.name, (0, 1, 2))
    with pytest.raises(KeyError):  # no schema
        schemas.parent_path("invalid", (0, 0, 1))

    # children
    expected = {b_ref}
    assert schemas.children(d_ref.name, d_ref.version) == expected
    assert schemas.children(d_ref) == {b_ref} == expected
    assert schemas.children(DirMeta) == {b_ref} == expected

    assert schemas.children(b_ref.name, b_ref.version) == set()

    with pytest.raises(KeyError):  # no such version of schema
        schemas.children(b_ref.name, (0, 1, 2))
    with pytest.raises(KeyError):  # no schema
        schemas.children("invalid", (0, 0, 1))



"""Helpers to create pydantic models that parse into and (de-)serialize to JSON-LD."""
from __future__ import annotations

from collections import ChainMap
from typing import Any, Dict, Mapping, Optional, TypeVar, Union

from pydantic import Extra, Field
from typing_extensions import Annotated, TypeAlias

from .core import MetadataSchema
from .decorators import add_const_fields
from .types import NonEmptyStr


def with_key_prefix(prefix: str, dct: Mapping[str, Any]) -> Dict[str, Any]:
    """Return new dict with all keys prefixed by `prefix`."""
    return {f"{prefix}{k}": v for k, v in dct.items() if v is not None}


def ld_decorator(**presets):
    """Return LD schema decorator with pre-set fields, e.g. `@context`.

    The returned decorator will attach the passed fields to a schema.

    All additional fields passed to the decorator will also be added,
    if not present, or override the default that is passed to this function.

    Note that the pre-set fields will ALWAYS override existing fields,
    regardless of the state of the override flag.

    Example usage:
    Pass your `@context` as `context` to this decorator factory.
    Use the returned decorator with `type` in order to
    set both the `@context` and `@type` for a schema.

    By default, will silently override any inherited
    constant fields that already exist in the schema.
    """

    def decorator(schema=None, *, override: bool = True, **kwargs):
        fields = with_key_prefix("@", ChainMap(kwargs, presets))
        dec = add_const_fields(fields, override=override)
        return dec if schema is None else dec(schema)

    return decorator


ld = ld_decorator()
"""Decorator to add constant JSON-LD fields equal for all instances of a schema."""


class LDSchema(MetadataSchema):
    """Semantically enriched schema for JSON-LD."""

    id_: Annotated[Optional[NonEmptyStr], Field(alias="@id")]

    def ref(self) -> LDIdRef:
        """Return LDIdRef, i.e. a pure @id reference for object.

        Throws an exception if no @id is found.
        """
        if self.id_ is None:
            raise ValueError("Object has no @id attribute!")
        return LDIdRef(id_=self.id_)

    @property
    def is_ld_ref(self):
        return False


class LDIdRef(LDSchema):
    """Object with just an @id reference (more info is given elsewhere)."""

    class Config:
        extra = Extra.forbid

    id_: Annotated[NonEmptyStr, Field(alias="@id")]

    def ref(self) -> LDIdRef:
        return self

    @property
    def is_ld_ref(self):
        return True


T = TypeVar("T", bound=LDSchema)

LDOrRef: TypeAlias = Union[LDIdRef, T]
"""LDOrRef[T] is either an object of LD Schema T, or a reference to an object.

An LD reference is just an object with an @id.
"""

LDRef: TypeAlias = LDIdRef  # Annotated[LDIdRef, T]  # <- does not work :(
"""LDRef[T] is a reference to an object of type T.

An LD reference is just an object with an @id.
"""



"""Test IH5MF record with manifest."""
from pathlib import Path
from uuid import uuid1

import pytest

from metador_core.ih5.manifest import IH5Manifest, IH5MFRecord, IH5UBExtManifest
from metador_core.ih5.record import IH5Record, IH5UserBlock


def latest_manifest_filepath(ds):
    return ds._manifest_filepath(ds._files[-1].filename)


def test_ubext_manifest():
    ub = IH5UserBlock.create()
    assert IH5UBExtManifest.get(ub) is None
    ubext = IH5UBExtManifest(
        is_stub_container=True, manifest_uuid=uuid1(), manifest_hashsum="sha256:0"
    )
    ubext.update(ub)
    assert IH5UBExtManifest.get(ub) == ubext


def test_fresh_record_no_manifest_exception(tmp_ds_path):
    with IH5MFRecord(tmp_ds_path, "w") as ds:
        with pytest.raises(ValueError):
            ds.manifest  # fresh uncommitted record has no manifest
        ds.commit_patch()
        ds.manifest  # no exception now


def test_commit_with_exts(tmp_ds_path):
    with IH5MFRecord(tmp_ds_path, "w") as ds:
        ds.commit_patch(manifest_exts={"test_ext": "yeah!"})
    # desired extra metadata section is stored successfully
    with IH5MFRecord(tmp_ds_path) as ds:
        assert ds.manifest.manifest_exts["test_ext"] == "yeah!"


def test_load_ih5mf_as_ih5(tmp_ds_path):
    # IH5MF containers (with patches and manifests) open cleanly as IH5 (ignore ext.)
    mf_files = []
    with IH5MFRecord(tmp_ds_path, "w") as ds:
        ds["foo/bar"] = "hello"
        ds.commit_patch()
        mf_files.append(latest_manifest_filepath(ds))

        ds.create_patch()
        ds["foo/baz"] = "world"
        ds.commit_patch()
        mf_files.append(latest_manifest_filepath(ds))

        assert len(ds.ih5_meta[-1].ub_exts.keys()) > 0  # userblock ext present

    # Manifest files exist for each patch
    assert all(map(lambda x: Path(x).is_file(), mf_files))

    with IH5Record(tmp_ds_path) as ds:
        # should make no problems, even though there are manifest files
        # and extras in the user block
        assert ds["foo/bar"][()] == b"hello"
        assert ds["foo/baz"][()] == b"world"


def test_commit_no_patch_fail(tmp_ds_path):
    # opening ih5mf with latest container lacking correct ub ext should fail
    with IH5MFRecord(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        with pytest.raises(ValueError):  # not writable
            ds.commit_patch()


def test_missing_ub_ext_fail(tmp_ds_path):
    # opening ih5mf with latest container lacking correct ub ext should fail
    with IH5MFRecord(tmp_ds_path, "w") as ds:
        pass
    with IH5Record(tmp_ds_path, "r+") as ds:
        pass  # commits patch without manifest!
    with IH5MFRecord(tmp_ds_path, "r+") as ds:
        with pytest.raises(ValueError):  # no manifest
            ds.manifest
        ds.commit_patch()
        ds.manifest  # now exists


def test_missing_latest_manifest_fail(tmp_ds_path):
    # opening ih5mf with latest container lacking manifest should fail
    ds_mf = ""
    with IH5MFRecord(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        ds.create_patch()
        ds["foo"] = 0
        ds.commit_patch()
        ds_mf = latest_manifest_filepath(ds)

    ds_mf.unlink()  # kill latest manifest
    with pytest.raises(ValueError):
        IH5MFRecord(tmp_ds_path)


def test_modified_latest_manifest_fail(tmp_ds_path):
    # opening ih5mf with latest container lacking manifest should fail
    ds_mf = ""
    with IH5MFRecord(tmp_ds_path, "w") as ds:
        ds.commit_patch()
        ds.create_patch()
        ds["foo"] = 0
        ds.commit_patch()
        ds_mf = latest_manifest_filepath(ds)

    with open(ds_mf, "w") as f:
        f.write("{}")  # overwrite manifest... changes hashsum

    with pytest.raises(ValueError):
        IH5MFRecord(tmp_ds_path)


def test_create_stub_from_mf_and_patch(tmp_ds_path_factory):
    # Skeleton from manifest can be used to create a valid stub and patch it
    ds1 = tmp_ds_path_factory()
    mf = ""
    files = []
    with IH5MFRecord(ds1, "w") as ds:
        ds["foo/bar"] = "hello"
        ds.commit_patch()

        ds.create_patch()
        ds["foo/baz"] = "world"
        ds.commit_patch()
        files = ds.ih5_files
        mf = latest_manifest_filepath(ds)

    # create a stub and add a patch
    ds2 = tmp_ds_path_factory()
    mf2 = ""
    with IH5MFRecord.create_stub(ds2, mf) as stub:
        stub.create_patch()
        stub["qux"] = "patch"
        stub.commit_patch()
        mf2 = latest_manifest_filepath(stub)
        files.append(stub.ih5_files[-1])

    # apply the patch to the orig dataset
    with IH5MFRecord._open(files, manifest_file=mf2) as ds:
        assert ds["qux"][()] == b"patch"


def test_merge_stub_fail(tmp_ds_path_factory):
    # Merging with a stub container fails
    ds_path = tmp_ds_path_factory()
    mf = ""
    with IH5MFRecord(ds_path, "w") as ds:
        ds["foo/bar"] = "hello"
        ds.commit_patch()
        mf = latest_manifest_filepath(ds)

    stub_path = tmp_ds_path_factory()
    merged_path = tmp_ds_path_factory()
    with IH5MFRecord.create_stub(stub_path, mf) as stub:
        assert IH5UBExtManifest.get(stub.ih5_meta[-1]).is_stub_container

        stub.create_patch()
        stub["qux"] = "patch"
        stub.commit_patch()

        with pytest.raises(ValueError) as e:
            stub.merge_files(merged_path)
        assert str(e).find("stub") >= 0


def test_merge_correct(tmp_ds_path_factory):
    # Merging should preserve latest manifest
    ds1_path = tmp_ds_path_factory()
    ds2_path = tmp_ds_path_factory()
    mf1_path = ""
    files = []
    # manifest extension names
    ext1 = "test_mfext"
    ext2 = "fresh_mfext"
    with IH5MFRecord(ds1_path, "w") as ds:
        ds["foo/bar"] = "hello"
        ds.commit_patch(manifest_exts={ext1: "hello"})
        mf_v1 = ds.manifest

        ds.create_patch()
        ds["foo/baz"] = "world"
        ds.manifest.manifest_exts[ext2] = "world"
        ds.commit_patch()
        mf_v2 = ds.manifest

        # new manifest should be fresh but inherit attached data
        assert mf_v1.manifest_uuid != mf_v2.manifest_uuid
        assert mf_v1.manifest_exts.get(ext1) == mf_v2.manifest_exts.get(ext1)
        assert mf_v2.manifest_exts.get(ext2) == "world"

        files = ds.ih5_files
        mf1_path = latest_manifest_filepath(ds)

        ds.merge_files(ds2_path)

    with IH5MFRecord(ds2_path, "r") as ds:
        # we should open this successfully... now check that the manifest agrees
        mf2_path = latest_manifest_filepath(ds)
        mf1 = IH5Manifest.parse_file(mf1_path)
        mf2 = IH5Manifest.parse_file(mf2_path)
        assert mf1 == mf2
        assert ext1 in mf2.manifest_exts  # also has the extensions
        assert ext2 in mf2.manifest_exts

    with IH5MFRecord(ds2_path, "r+") as ds:
        # create a new patch on top of merged container
        ds["qux"] = 123
        files.append(ds.ih5_files[-1])

    # open patch done on top of merged container, now with the original containers.
    # it should infer manifest file of most recent patch and work fine.
    with IH5MFRecord._open(files) as ds:
        assert ds["qux"][()] == 123
        assert "test_mfext" in ds.manifest.manifest_exts
